/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _ADDITIVE_SEMIGROUP_HPP
#define _ADDITIVE_SEMIGROUP_HPP

#include <boost/mpl/vector.hpp>
#include "semigroup.hpp"
#include "map.hpp"
#include "set.hpp"
#include "root.hpp"

//arithmetic with +

template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Add;
//the next three classes should not be specialiced.
//should be templated typedefs
template<typename S> struct Zero;
template<typename S> struct EqualsZero;
template<typename S> struct NegatedElement;
template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Sub;

//metafunction
template<typename S> struct AdditiveSemigroup {
    typedef Semigroup<Add<S> > type;
};

template<typename S>
struct Add<Root, S, TypeConversionByImplementation> : public Map<boost::mpl::vector<S, S>, S> {
    Element<S> operator()(S const & a, S const & b);
};

template<typename S>
struct Sub<Root, S, TypeConversionByImplementation> : public Map<boost::mpl::vector<S, S>, S> {
    Element<S> operator()(Element<S> const & a, Element<S> const & b) const {
	return Add<S>()(a,NegatedElement<S>()(b));
    }
};

template<typename S_base, typename S, typename TC> struct Add :
    public Add<
	typename Base<S_base, Add<Root> >::type,
	S,
	typename TypeConversion<S_base, Add<Root> >::type
    >
{};

template<typename S> struct Zero : public NeutralElement<Semigroup<Add<S> > > {};
template<typename S> struct EqualsZero : public EqualsNeutralElement<Semigroup<Add<S> > > {};
template<typename S> struct NegatedElement : public InverseElement<Semigroup<Add<S> > > {};

template<typename S_base, typename S, typename TC> struct Sub :
    public Sub<
	typename Base<S_base, Sub<Root> >::type,
	S,
	typename TypeConversion<S, Sub<Root> >::type
    >
{};

template<typename S>
Element<S> zero() {
    return Zero<S>()();
}

template<typename S>
bool equalsZero(Element<S> const & a) {
    return EqualsZero<S>()(a);
}

template<typename S>
Element<S> negatedElement(Element<S> const & a) {
    return NegatedElement<S>()(a);
}

template<typename S>
Element<S> operator+(Element<S> const & a, Element<S> const & b) {
    return Add<S>()(a,b);
}

template<typename S>
Element<S> operator-(Element<S> const & a) {
    return NegatedElement<S>()(a);
}

template<typename S>
Element<S> operator-(Element<S> const & a, Element<S> const & b) {
    return Sub<S>()(a,b);
}

template<typename S>
Element<S> & operator+=(Element<S> & a, Element<S> const & b) {
//there is no counterpart for "InLeftPlace", so be careful for non-commutative Add!
//Furthermore, see the inconsistencies for "InLeftPlace" in binaryOperation.hpp
//Whenever possible, use InRightPlace.
    return InRightPlace<Semigroup<Add<S> > >()(b,a);
}

template<typename S>
Element<S> & operator-=(Element<S> & a, Element<S> const & b) {
    return a+=(-b);
}
#endif
