/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _ALGORITHM_HPP
#define _ALGORITHM_HPP

#include <boost/mpl/bool.hpp>
#include "hashHelper.hpp"

// RepeatedApplicationReaches

template<typename T, typename Func, typename GoalComp, typename NeedsNoHash>
struct RepeatedApplicationReaches {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(T const & a, Func func, GoalComp goalComp, NeedsNoHash needsNoHash) const;
};

template<typename T, typename Func, typename GoalComp>
struct RepeatedApplicationReaches<T,Func,GoalComp,boost::mpl::false_> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(T const & a, Func func, GoalComp goalComp) const {
	HASH_SET(T) visited;
	for(T b = a; ; b = func(b)) {
	    if(goalComp(b)) return true;
	    if(visited.count(b)) return false;
	    visited.insert(b);
	}
    }
};

template<typename T, typename Func, typename GoalComp>
struct RepeatedApplicationReaches<T,Func,GoalComp,boost::mpl::true_> {
    template <typename Args> struct sig {
	typedef bool type;
    };
    bool operator()(T const & a, Func func, GoalComp goalComp) const {
	T b = a;
	while (true) {
	    if(goalComp(b)) return true;
	    b = func(b);
	    if(b == a) return false;
	}
    }
};

template<typename T, typename Func, typename GoalComp, typename NeedsNoHash>
bool repeatedApplicationReaches(T const & a, Func func, GoalComp goalComp, NeedsNoHash needsNoHash) {
    return RepeatedApplicationReaches<T, Func, GoalComp, NeedsNoHash>()(a, func, goalComp);
}

// RepeatedApplicationIndex
template<typename T, typename Func, typename GoalComp, typename SkipCheck, typename NeedsNoHash = SkipCheck>
struct RepeatedApplicationIndex {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(T const & a) const;
    /* returns 0 if the repeated application of 'Func' on 'a' never fulfills GoalComp,
     * or 1 + the number of applications of 'Func' after which 'a' fulfills GoalComp the first time.
     * If it is known that the first case will not happen, SkipCheck can be set to true.
     */
};

template<typename T, typename Func, typename GoalComp>
struct RepeatedApplicationIndex<T,Func,GoalComp,boost::mpl::false_,boost::mpl::false_> {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(T const & a, Func func, GoalComp goalComp) const {
	unsigned int res = 1;
	HASH_SET(T) visited;
	for(T b = a; ; b = func(b)) {
	    if(goalComp(b)) return res;
	    if(visited.count(b)) return 0;
	    visited.insert(b);
	    ++res;
	}
    }
};

template<typename T, typename Func, typename GoalComp>
struct RepeatedApplicationIndex<T,Func,GoalComp,boost::mpl::false_,boost::mpl::true_> {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(T const & a, Func func, GoalComp goalComp) const {
	unsigned int res = 1;
	T b = a;
	while (true) {
	    if(goalComp(b)) return res;
	    b = func(b);
	    if(b == a) return 0;
	    ++res;
	}
    }
};

template<typename T, typename Func, typename GoalComp, typename SkipCheck>
struct RepeatedApplicationIndex<T,Func,GoalComp,boost::mpl::true_,SkipCheck> {
    template <typename Args> struct sig {
	typedef unsigned int type;
    };
    unsigned int operator()(T const & a, Func func, GoalComp goalComp) const {
	unsigned int res = 1;
	T b = a;
	while (true) {
	    if(goalComp(b)) return res;
	    b = func(b);
	    ++res;
	}
    }
};

template<typename T, typename Func, typename GoalComp, typename SkipCheck, typename NeedsNoHash> 
unsigned int repeatedApplicationIndex(T const & a, Func func, GoalComp goalComp, SkipCheck skipCheck, NeedsNoHash needsNoHash) {
    return RepeatedApplicationIndex<T,Func,GoalComp,SkipCheck,NeedsNoHash>()(a,func,goalComp);
}
#endif
