/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _AUTOMORPHISM_GROUP_HPP
#define _AUTOMORPHISM_GROUP_HPP

//the Automorphism Group of some Structure, for Example of Ring, RingWithIdentity, ...

#include "root.hpp"
#include "set.hpp"
#include "action.hpp"
#include "semigroup.hpp"
#include "multiplicativeSemigroup.hpp"
#include "map.hpp"
#include <boost/mpl/vector.hpp>
#include <iostream>

template<typename Struct_base, typename Struct = Struct_base>
struct AutomorphismGroup_BaseSet;

template<typename Struct>
struct AutomorphismGroup_BaseSet<Root, Struct> {
    typedef Struct BaseStructure;
};

template<typename Struct_base, typename Struct>
struct AutomorphismGroup_BaseSet : public AutomorphismGroup_BaseSet<Root, Struct> {};

template<typename Struct_base, typename Struct = Struct_base>
struct AutomorphismGroup_Operator;

template<typename Struct>
struct AutomorphismGroup_Operator<Root, Struct> : public Map<boost::mpl::vector<AutomorphismGroup_BaseSet<Struct>,AutomorphismGroup_BaseSet<Struct> >,AutomorphismGroup_BaseSet<Struct> >
{
    typedef Struct BaseStructure;
};

template<typename Struct_base, typename Struct>
struct AutomorphismGroup_Operator : public AutomorphismGroup_Operator<Root, Struct> {};

template<typename Struct>
struct AutomorphismGroup {
    typedef Semigroup_Atom<AutomorphismGroup_Operator<Struct> > type;
};

template<typename AutGrp_base, typename AutGrp = AutGrp_base> class Application;

template<typename AutGrp>
class Application<Root, AutGrp> {
    private:
	typedef typename AutGrp::type::BaseSet X_;
    public:
        template <typename Args> struct sig {
            typedef Element<X_> & type;
        };
        typedef Element<X_> & result_type;
	Element<X_> operator()(Element<AutGrp> const & aut, Element<X_> const & x) const;
};

template<typename AutGrp_base, typename AutGrp>
struct Application :
    public Application<
        typename Base<AutGrp_base, Application<Root> >::type,
        AutGrp
    >
{};

template<typename Struct> class AutomorphismAction {
    typedef Action_Atom<AutomorphismGroup_Operator<Struct>, Application<AutomorphismGroup<Struct> > > type;
};

template<typename Struct, typename Derived>
class NameStream<AutomorphismGroup_BaseSet<Struct>, Derived> : public NameStream<Root, Derived> {
    private:
	typedef typename Derived::type::BaseSet X_;

    public:
	std::string & operator()(std::string & o) const {
	    return "Aut(" << NameStream<X_>()() << ")";
	}
};

template<typename S>
struct HasNeutralElement<AutomorphismGroup<S> > : public boost::mpl::bool_<true> {};

template<typename S>
struct HasInverseElements<AutomorphismGroup<S> > : public boost::mpl::bool_<true> {};

template<typename S>
struct IsCommutative<AutomorphismGroup<S> > : public boost::mpl::bool_<false> {};

#endif
