/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _CARTESIAN_PRODUCT_SET_HPP
#define _CARTESIAN_PRODUCT_SET_HPP

#include <iostream>
#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/iterator/iterator_facade.hpp>
#include "set.hpp"
#include "mathHelper.hpp"
#include "root.hpp"

//cartesianPowerSet.hpp
template<typename S, typename T> class CartesianProductSet;
template<typename CP_base, typename CP = CP_base, typename TC = TypeConversionByImplementation> struct Left;
template<typename CP_base, typename CP = CP_base, typename TC = TypeConversionByImplementation> struct Right;
template<typename CP_base, typename CP = CP_base, typename TC = TypeConversionByImplementation> struct SetLeft;
template<typename CP_base, typename CP = CP_base, typename TC = TypeConversionByImplementation> struct SetRight;

template<typename S, typename T> class Element<CartesianProductSet<S, T> >;

template<typename S, typename T>
class CartesianProductSet {
    public:
	typedef S LeftBaseSet;
	typedef T RightBaseSet;
	typedef std::pair<S, T> DataV_;
};

template<typename S, typename T>
class Element<CartesianProductSet<S, T> > : public Element_root<CartesianProductSet<S, T> > {
    private:
	Element<S> leftValue;
	Element<T> rightValue;

    public:
	typedef CartesianProductSet<S, T> CP_;

	Element() : leftValue(Element<S>()), rightValue(Element<T>()) {}
	Element(Element<S> const & leftValueIn, Element<T> const & rightValueIn) : leftValue(leftValueIn), rightValue(rightValueIn) {}
	Element(std::pair<Element<S>,Element<T> > const & pairIn) : leftValue(pairIn.first), rightValue(pairIn.right) {}

	Element<S> const & left() const {
	    return leftValue;
	}

	Element<S> & left() {
	    return leftValue;
	}

	Element<T> const & right() const {
	    return rightValue;
	}

	Element<T> & right() {
	    return rightValue;
	}

	std::pair<Element<S>,Element<T> > pair() const {
	    return std::pair<Element<S>,Element<T> >(leftValue,rightValue);
	}

	void set(Element<S> const & leftValueIn, Element<T> const & rightValueIn) {
	    leftValue = leftValueIn;
	    rightValue = rightValueIn;
	}

	std::pair<Element<S>,Element<T> > & set(std::pair<Element<S>,Element<T> > const pairIn) {
	    leftValue = pairIn.first;
	    rightValue = pairIn.second;
	    return pairIn;
	}
};

//template<typename CP>
//struct CartesianProductTypedefs {
//    protected:
//	typedef CP CP_;
//	typedef typename CP_::LeftBaseSet CP_L_;
//	typedef typename CP_::RightBaseSet CP_R_;
//};

template<typename CP, typename TC>
struct Left<Root, CP, TC> {
    template <typename Args> struct sig {
        typedef Element<typename CP::LeftBaseSet> & type;
    };
    Element<typename CP::LeftBaseSet> & operator()(Element<CP> & a) const {
	return a.left();
    }
    Element<typename CP::LeftBaseSet> const & operator()(Element<CP> const & a) const {
	return a.left();
    }
};

template<typename CP, typename TC>
struct Right<Root, CP, TC> {
    template <typename Args> struct sig {
        typedef Element<typename CP::RightBaseSet> & type;
    };
    Element<typename CP::RightBaseSet> & operator()(Element<CP> & a) const {
	return a.right();
    }
    Element<typename CP::RightBaseSet> const & operator()(Element<CP> const & a) const {
	return a.right();
    }
};

template<typename CP, typename TC>
struct SetLeft<Root, CP, TC> {
    template <typename Args> struct sig {
        typedef Element<typename CP::LeftBaseSet> & type;
    };
    Element<typename CP::LeftBaseSet> & operator()(Element<CP> & a, Element<typename CP::LeftBaseSet> leftIn) const {
	return a.left() = leftIn;
    }
};

template<typename CP, typename TC>
struct SetRight<Root, CP, TC> {
    template <typename Args> struct sig {
        typedef Element<typename CP::RightBaseSet> & type;
    };
    Element<typename CP::RightBaseSet> & operator()(Element<CP> & a, Element<typename CP::RightBaseSet> rightIn) const {
	return a.right() = (rightIn);
    }
};

template<typename S_base, typename S, typename TC> struct Left :
    public Left<
        typename Base<S_base, Left<Root> >::type,
        S,
        typename TypeConversion<S_base, Left<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct Right :
    public Right<
        typename Base<S_base, Right<Root> >::type,
        S,
        typename TypeConversion<S_base, Right<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct SetLeft :
    public SetLeft<
        typename Base<S_base, SetLeft<Root> >::type,
        S,
        typename TypeConversion<S_base, SetLeft<Root> >::type
    >
{};

template<typename S_base, typename S, typename TC> struct SetRight :
    public SetRight<
        typename Base<S_base, SetRight<Root> >::type,
        S,
        typename TypeConversion<S_base, SetRight<Root> >::type
    >
{};


// set.hpp
template<typename S, typename T, typename Derived>
class Equals<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public Equals<Root, Derived> {
//class Equals<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public Equals<Root, Derived>, public CartesianProductTypedefs<Derived> {
    public:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
	bool operator()(Element<CP_> const & v, Element<CP_> const & w) {
	    return ((v.left() == w.left()) && (v.right() == w.right()));
	}
};

template<typename S, typename T, typename Derived>
class NameStream<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public NameStream<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
    public:
	std::ostream & operator()(std::ostream & o) {
	    return NameStream<T>()(NameStream<S>()(o) << " x ");
	}
};

template<typename S, typename T, typename Derived>
class ElementName<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public ElementName<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
    public:
    std::ostream & operator()(std::ostream & o, Element<CP_> const & v) const {
	o << "[ " << v.left() << ", " << v.right() << " ]";
	return o;
    }
};

template<typename S, typename T, typename Derived>
class HashValue<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public HashValue<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
    public:
	size_t operator()(Element<CP_> const & v) {
	    //not optimal:
	    return HashValue<CP_L_>()(v.left()) * Cardinality<CP_R_>()() + HashValue<CP_R_>()(v.right());
	}
};

template<typename S, typename T, typename Derived>
class RankElement<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public RankElement<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
    public:
	unsigned int operator()(Element<CP_> const & v) {
	    return RankElement<CP_L_>()(v.left()) * Cardinality<CP_R_>()() + RankElement<CP_R_>()(v.right());
	}
};

template<typename S, typename T, typename Derived>
class UnrankElement<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public UnrankElement<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
    public:
	Element<CP_> operator()(unsigned int const & n) {
	    unsigned int card = Cardinality<CP_R_>()();
	    return Element<CP_>(UnrankElement<CP_L_>()(n / card), UnrankElement<CP_R_>()(n % card));
	}
};

template<typename S, typename T, typename Derived>
class Compare<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public Compare<Root, Derived> {
    //lexicographic order
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;
    public:
	bool operator()(Element<CP_> const & v, Element<CP_> const & w) {
	    return (Compare<CP_L_>()(v.left(),w.left()) || ((v.left() == w.left()) && Compare<CP_R_>()(v.right(), w.right())));
	}
};

template<typename S, typename T, typename Derived>
class Cardinality<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public Cardinality<Root, Derived> {
    //only works if it fits into a unsigned int
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;

    public:
	unsigned int operator()() {
	    return Cardinality<CP_L_>()() * Cardinality<CP_R_>()();
	}
};

template<typename S, typename T, typename Derived>
struct Random<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public Random<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;

    public:
	Element<CP_> operator()() const {
	    return Element<CP_>(Random<CP_L_>()(),Random<CP_R_>()());
	}
};

template<typename S, typename T, typename Derived>
struct Begin<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation>;

template<typename S, typename T, typename Derived>
struct End<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation>;

template<typename S, typename T, typename Derived>
struct IsAtEnd<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation>;

template<typename S, typename T, typename Derived>
struct Iterator<CartesianProductSet<S, T>, Derived> : 
public boost::iterator_facade<
    Iterator<CartesianProductSet<S, T>, Derived>,
    Element<Derived> const,
    boost::incrementable_traversal_tag
>
{
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;

    friend class boost::iterator_core_access;
    friend struct Begin<CartesianProductSet<S, T>, Derived>;
    friend struct End<CartesianProductSet<S, T>, Derived>;
    //friend struct IsAtBegin<CartesianProductSet<S, T>, Derived>;
    friend struct IsAtEnd<CartesianProductSet<S, T>, Derived>;

    public:
	Iterator() {}

    private:
	Iterator<CP_L_> outer;
	Iterator<CP_R_> inner;
	Element<CP_> value;
	bool atEnd;

	void increment() {
	    ++inner;
	    if (IsAtEnd<CP_R_>()(inner)) {
		++outer;
		if (IsAtEnd<CP_L_>()(outer)) {
		    atEnd = true;
		    return;
		}
		inner = Begin<CP_R_>()();
		value.left() = (*outer);
	    }
	    value.right() = (*inner);
	}

	bool equal(Iterator<CartesianProductSet<S, T>, Derived> const & other) const {
	    return (value == other.value && atEnd == other.atEnd);
	}

	Element<CP_> const & dereference() const {
	    return value;
	}
};

template<typename S, typename T, typename Derived>
class Begin<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public Begin<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;

    public:
	Iterator<CP_> operator()() const {
	    Iterator<CP_> res;
	    res.outer = Begin<CP_L_>()();
	    res.inner = Begin<CP_R_>()();
	    res.atEnd = false;
	    res.value.set(*(res.outer),*(res.inner));
	    return res;
	}
};

/*
//CAUTION: activation of end needs an adaption of Iterator::equal()
template<typename S, typename T, typename Derived>
class End<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public End<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;

    public:
	Iterator<CP_> operator()() const {
	    Iterator<CP_> result;
	    result.atEnd = true;
	    return result;
	}
};
*/

template<typename S, typename T, typename Derived>
class IsAtEnd<CartesianProductSet<S, T>, Derived, TypeConversionByImplementation> : public IsAtEnd<Root, Derived> {
    private:
	typedef Derived CP_;
	typedef typename CP_::LeftBaseSet CP_L_;
	typedef typename CP_::RightBaseSet CP_R_;

    public:
	bool operator()(Iterator<CP_> const & it) const {
	    return it.atEnd;
	}
};
#endif
