/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _HASH_HELPER_HPP
#define _HASH_HELPER_HPP

#include "set.hpp"

#ifdef GNU
    #include <tr1/unordered_set>
    #include <tr1/unordered_map>
    #define HASH_SET(T) std::tr1::unordered_set<T>
    #define HASH_MAP(K,V) std::tr1::unordered_map<K,V>
#else
    #include <hash_set>
    #include <hash_map>
    #define HASH_SET(T) stdext::hash_set<T,hashClass<T> >
    #define HASH_MAP(K,V) stdext::hash_map<K,V,hashClass<K> >
#endif

//wrapper function object for hashing, depending on the standard library
#ifdef GNU
//for gnu standard library
namespace std {
    namespace tr1 {
	template<typename S>
	    struct hash<Element<S> > : std::unary_function<Element<S>, size_t>{
		size_t operator()(const Element<S> & a) const {
		    return HashValue<S>()(a);
		}
	    };
    }
}

#else
//for Micro$oft standard library
template <typename E>
class hashClass : public stdext::hash_compare<E>
{
public:
	size_t operator() (const E& a) const {
		return HashValue<typename E::BaseSet>()(a);
	}

	bool operator() (const  E& a, const E& b) const {
		return Compare<typename E::BaseSet>()(a,b);
	}
};
#endif
#endif

/*
class HashLongLong
{
    public:
	size_t operator()(const unsigned long long ld) const
	{
	    return (size_t)ld;
	}
};


class EqualLongLong
{
    public:
	bool operator()(const unsigned long long ld1, const unsigned long long ld2)
	    const
	    { return ld1 == ld2; }
};

typedef __gnu_cxx::hash_set<unsigned long long, HashLongLong, EqualLongLong>
LongLongHashSet;
*/
