/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef MATRIX_MUL_HPP
#define MATRIX_MUL_HPP

#include "matrixSet.hpp"
#include "map.hpp"
#include "ring.hpp"
#include "additiveSemigroup.hpp"
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/mpl/vector.hpp>
// * implementations *

template<typename Mres, typename Mleft, typename Mright, typename R>
struct MatrixMul_impl : public Map<boost::mpl::vector<Mleft,Mright>,Mres> {
    Element<Mres> operator()(Element<Mleft> const & a, Element<Mright> const & b) const {
	BOOST_STATIC_ASSERT((boost::is_same<typename Mres::BaseSet, typename Mleft::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Mres::BaseSet, typename Mright::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Mres::BaseSet, typename R::type::BaseSet>::value));
	//assert(NumberOfRows<Mres>()() == NumberOfRows<Mleft>()());
	//assert(NumberOfColumns<Mres>()() == NumberOfColumns<Mright>()());
	//assert(NumberOfColumns<Mleft>()() == NumberOfRows<Mright>()());
	//assert(NumberOfRows<Mres>()() != 0);
	//assert(NumberOfColumns<Mres>()() != 0);
	//assert(NumberOfRows<Mright>()() != 0);

	typedef R BaseRing;
	typedef typename R::type::BaseSet R_BS_;
	typedef typename R::type::Mul R_Mul;
	typedef typename R::type::Add R_Add;
	typedef typename std::vector<Element<R_BS_> >::iterator It;
	typedef typename std::vector<Element<R_BS_> >::const_iterator Itc;
	Itc leftit,rightit,lit,rit,litub;
	Element<Mres> res;
	It resit = res.data.begin();
	for(leftit = a.data.begin(); leftit != a.data.end(); leftit += NumberOfColumns<Mres>()()) {
	    for(rightit = b.data.begin(); rightit != b.data.begin() + NumberOfColumns<Mres>()(); ++rightit) {
		lit = leftit;
		litub = leftit + NumberOfRows<Mright>()();
		rit = rightit;
		(*resit) = R_Mul()((*lit),(*rit));
loop_Mul_MatrixMul_impl:
		++lit;
		if(lit != litub) {
		    rit += NumberOfColumns<Mright>()();

		    InRightPlace<Semigroup<R_Add> >()(R_Mul()((*lit),(*rit)),*resit);
		    goto loop_Mul_MatrixMul_impl;
		}
		++resit;
	    }
	}
	return res;
    }
};

template<typename Vres, typename Vleft, typename Mright, typename R>
struct VectorMatrixMul_impl : public Map<boost::mpl::vector<Mright, Vleft>,Vres> {
    Element<Vres> operator()(Element<Mright> const & m, Element<Vleft> const & v) const {
	//CAUTION: syntax is consistent with action.hpp!
	//so VectorMatrixMul_impl()(m,v) means v*m.
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename Vleft::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename Mright::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename R::type::BaseSet>::value));
	//assert(Exponent<Vres>()() == NumberOfRows<Vleft>()());
	//assert(NumberOfColumns<Vres>()() == NumberOfColumns<Mright>()());
	//assert(NumberOfColumns<Vleft>()() == Exponent<Mright>()());
	//assert(Exponent<Vres>()() != 0);
	//assert(NumberOfColumns<Vres>()() != 0);
	//assert(Exponent<Mright>()() != 0);
	Element<Vres> res;

	typedef R BaseRing;
	typedef typename R::type::BaseSet R_BS_;
	typedef typename R::type::Mul R_Mul;
	typedef typename R::type::Add R_Add;
	typedef typename std::vector<Element<R_BS_> >::iterator It;
	typedef typename std::vector<Element<R_BS_> >::const_iterator Itc;
	Itc vit = v.data.begin();
	Itc mit = m.data.begin();
	Itc const vitub = v.data.end();
	Itc const resub = res.data.end();
	It resit = res.data.begin();

	for( ; resit != resub; ++resit) {
	    (*resit) = R_Mul()((*vit),(*mit)); 
	    ++mit;
	}
loop_Mul_VectorMatrixMul_impl:
	++vit;
	if (vit != vitub) {
	    for(resit = res.data.begin(); resit != resub; ++resit) {
		InRightPlace<Semigroup<R_Add> >()(R_Mul()((*vit),(*mit)),*resit);
		++mit;
	    }
	    goto loop_Mul_VectorMatrixMul_impl;
	}
	return res;
    }
};

template<typename Vres, typename Mleft, typename Vright, typename R>
struct MatrixVectorMul_impl : public Map<boost::mpl::vector<Mleft,Vright>,Vres> {
    Element<Vres> operator()(Element<Mleft> const & m, Element<Vright> const & v) const {
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename Mleft::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename Vright::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename R::type::BaseSet>::value));
	//assert(Exponent<Vres>()() == NumberOfRows<Mleft>()());
	//assert(NumberOfColumns<Vres>()() == NumberOfColumns<Vright>()());
	//assert(NumberOfColumns<Mleft>()() == Exponent<Vright>()());
	//assert(Exponent<Vres>()() != 0);
	//assert(NumberOfColumns<Vres>()() != 0);
	//assert(Exponent<Vright>()() != 0);
	Element<Vres> res;

	typedef R BaseRing;
	typedef typename R::type::BaseSet R_BS_;
	typedef typename R::type::Mul R_Mul;
	typedef typename R::type::Add R_Add;
	typedef typename std::vector<Element<R_BS_> >::iterator It;
	typedef typename std::vector<Element<R_BS_> >::const_iterator Itc;
	Itc vit,resit;
	Itc mit = m.data.begin();
	Itc const vitub = v.data.end();
	Itc const resub = res.data.end();
	for(It resit = res.data.begin(); resit != resub; ++resit) {
	    vit = v.data.begin();
	    (*resit) = R_Mul()((*mit),(*vit)); 
loop_Mul_MatrixVectorMul_impl:
	    ++mit;
	    ++vit;
	    if (vit != vitub) {
		InRightPlace<Semigroup<R_Add> >()(R_Mul()((*mit),(*vit)),*resit);
		goto loop_Mul_MatrixVectorMul_impl;
	    }
	}
	return res;
    }
};

template<typename Vres, typename Mleft, typename Vright, typename R>
struct ReverseMatrixVectorMul_impl : public Map<boost::mpl::vector<Mleft,Vright>,Vres> {
    Element<Vres> operator()(Element<Mleft> const & m, Element<Vright> const & v) const {
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename Mleft::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename Vright::BaseSet>::value));
	BOOST_STATIC_ASSERT((boost::is_same<typename Vres::BaseSet, typename R::type::BaseSet>::value));
	//assert(Exponent<Vres>()() == NumberOfRows<Mleft>()());
	//assert(NumberOfColumns<Vres>()() == NumberOfColumns<Vright>()());
	//assert(NumberOfColumns<Mleft>()() == Exponent<Vright>()());
	//assert(Exponent<Vres>()() != 0);
	//assert(NumberOfColumns<Vres>()() != 0);
	//assert(Exponent<Vright>()() != 0);
	Element<Vres> res;

	typedef R BaseRing;
	typedef typename R::type::BaseSet R_BS_;
	//typedef typename Opposite<Semigroup<typename R::type::Mul> >::type::Operator R_Mul;
	typedef typename R::type::Mul R_Mul;
	typedef typename R::type::Add R_Add;
	typedef typename std::vector<Element<R_BS_> >::iterator It;
	typedef typename std::vector<Element<R_BS_> >::const_iterator Itc;
	Itc vit,resit;
	Itc mit = m.data.begin();
	Itc const vitub = v.data.end();
	Itc const resub = res.data.end();
	for(It resit = res.data.begin(); resit != resub; ++resit) {
	    vit = v.data.begin();
	    //(*resit) = R_Mul()((*mit),(*vit)); 
	    (*resit) = R_Mul()((*vit),(*mit)); 
loop_Mul_MatrixVectorMul_impl:
	    ++mit;
	    ++vit;
	    if (vit != vitub) {
		//InRightPlace<Semigroup<R_Add> >()(R_Mul()((*mit),(*vit)),*resit);
		InRightPlace<Semigroup<R_Add> >()(R_Mul()((*vit),(*mit)),*resit);
		goto loop_Mul_MatrixVectorMul_impl;
	    }
	}
	return res;
    }
};


template<typename MSet,typename R>
struct MatrixMulSemigroup {
    typedef Semigroup_Atom<MatrixMul_impl<MSet,MSet,MSet,R> > type;
};

template<typename MSet, typename R>
struct Base<MatrixMulSemigroup<MSet, R> > {
    typedef Semigroup<MatrixMul_impl<MSet,MSet,MSet,R> > type;
};

template<typename MMSG_base, typename MMSG = MMSG_base> struct MatrixMulBaseRing;

template<typename MSet, typename R, typename Derived>
struct MatrixMulBaseRing<Semigroup<MatrixMul_impl<MSet,MSet,MSet,R> >, Derived> {
    typedef R type;
};

template<typename MMSG_base, typename MMSG> struct MatrixMulBaseRing :
    public MatrixMulBaseRing<
	typename Base<MMSG_base, MatrixMulBaseRing<Root> >::type,
	MMSG
    >
{};

template<typename MSet, typename R, typename Derived>
struct HasNeutralElement<Semigroup<MatrixMul_impl<MSet,MSet,MSet,R> >, Derived> : public HasNeutralElement<Semigroup<typename MatrixMulBaseRing<Derived>::type::type::Mul> > {};

template<typename MSet, typename R, typename Derived>
struct NeutralElement<Semigroup<MatrixMul_impl<MSet,MSet,MSet,R> >, Derived> : public NeutralElement<Root, Derived> { 
    typedef Derived MMS_;
    typedef typename MatrixMulBaseRing<MMS_>::type R_;
    typedef typename MMS_::type::Operator MM_Op;
    typedef typename R_::type::BaseSet RS_;
    typedef typename MMS_::type::BaseSet Mat_;
    typedef typename std::vector<Element<RS_> >::iterator Itc;
    Element<Mat_> operator()() const {
	Element<Mat_> res;
	Itc i = res.data.begin();
	Itc iend = res.data.end();
loop_NeutralElement_MatrixAlgebra:
	(*i) = NeutralElement<Semigroup<typename R_::type::Mul> >()();
	++i;
	if (i == iend) return res;
	for (unsigned int j = NumberOfColumns<Mat_>()(); j-->0;) {
	    (*i) = NeutralElement<Semigroup<typename R_::type::Add> >()();
	    ++i;
	}
	goto loop_NeutralElement_MatrixAlgebra;
    }
};

template<typename MSet, typename R, typename Derived>
struct EqualsNeutralElement<Semigroup<MatrixMul_impl<MSet,MSet,MSet,R> >, Derived> : public EqualsNeutralElement<Root, Derived> { 
    typedef Derived MMS_;
    typedef typename MatrixMulBaseRing<MMS_>::type R_;
    typedef typename MMS_::type::Operator MM_Op;
    typedef typename R_::type::BaseSet RS_;
    typedef typename MMS_::type::BaseSet Mat_;
    typedef typename std::vector<Element<RS_> >::const_iterator Itc;
    bool operator()(Element<Mat_> const & a) const {
	Itc i = a.data.begin();
	Itc iend = a.data.end();
loop_EqualsNeutralElement_MatrixAlgebra:
	if (!EqualsNeutralElement<Semigroup<typename R_::type::Mul> >()(*i)) return false;
	++i;
	if (i == iend) return true;
	for (unsigned int j = NumberOfColumns<Mat_>()(); j-->0;) {
	    if (!EqualsNeutralElement<Semigroup<typename R_::type::Add> >()(*i)) return false;
	    ++i;
	}
	goto loop_EqualsNeutralElement_MatrixAlgebra;
    }
};

template<typename R, typename id, typename Derived>
struct Mul<SquareMatrixSet<R, id>, Derived> : public MatrixMulSemigroup<Derived, Ring<Add<typename Derived::BaseSet>, Mul<typename Derived::BaseSet> > >::type::Operator {};

#endif
