/********************************************
Copyright 2010 by Michael Kiermaier, Bayreuth
********************************************/

#ifndef _MULTIPLICATIVE_SEMIGROUP_HPP
#define _MULTIPLICATIVE_SEMIGROUP_HPP

#include <boost/mpl/vector.hpp>
#include "semigroup.hpp"
#include "map.hpp"
#include "set.hpp"
#include "root.hpp"

//arithmetic with *

template<typename S_base, typename S = S_base, typename TC = TypeConversionByImplementation> struct Mul;
//dont specialize the next three classes.
template<typename S> struct One;
template<typename S> struct EqualsOne;
template<typename S> struct Inverse;

//metafunction
template<typename S> struct MultiplicativeSemigroup {
    typedef Semigroup<Mul<S> > type;
};

template<typename S>
struct Mul<Root, S, TypeConversionByImplementation> : public Map<boost::mpl::vector<S, S>, S> {
    Element<S> operator()(S const & a, S const & b);
};

template<typename S_base, typename S, typename TC> struct Mul :
    public Mul<
	typename Base<S_base, Mul<Root> >::type,
	S,
	typename TypeConversion<S_base, Mul<Root> >::type
    >
{};
template<typename S> struct One : public NeutralElement<Semigroup<Mul<S> > > {};
template<typename S> struct EqualsOne : public EqualsNeutralElement<Semigroup<Mul<S> > > {};
template<typename S> struct Inverse : public InverseElement<Semigroup<Mul<S> > > {};


template<typename S>
Element<S> one() {
    return One<S>()();
}

template<typename S>
bool equalsOne(Element<S> const & a) {
    return EqualsOne<S>()(a);
}

template<typename S>
Element<S> inverse(Element<S> const & a) {
    return Inverse<S>()(a);
}

template<typename S>
bool isInvertible(Element<S> const & a) {
    return IsInvertibleElement<Mul<S> >()(a);
}

template<typename S>
Element<S> operator*(Element<S> const & a, Element<S> const & b) {
    return Mul<S>()(a,b);
}

template<typename S>
Element<S> & operator*=(Element<S> & a, Element<S> const & b) {
//there is no counterpart for "InLeftPlace", so be careful for non-commutative Mul!
//Furthermore, see the inconsistencies for "InLeftPlace" in binaryOperation.hpp
//Whenever possible, use InRightPlace.
    return InRightPlace<Semigroup<Mul<S> > >()(b,a);
}

template<typename S>
Element<S> power(Element<S> const & a, unsigned int const expIn) {
    return Power<Semigroup<Mul<S> > >()(a,expIn);
}
#endif
