/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _PERMUTATION_GROUP_HPP
#define _PERMUTATION_GROUP_HPP

//#include <iostream>
//#include <fstream>
//#include <cassert>
//#include <string>
#include <vector>
#include <algorithm>
//#include <boost/mpl/bool.hpp>
//#include <boost/mpl/vector.hpp>
//#include <boost/iterator/iterator_facade.hpp>
//#include "mathHelper.hpp"
#include "numberRangeSet.hpp"
#include "valueWrapperPolicy.hpp"
#include "cartesianPowerSet.hpp"
//#include "action.hpp"
#include "semigroup.hpp"
#include "multiplicativeSemigroup.hpp"
//#include "set.hpp"
#include "root.hpp"
#include "map.hpp"
//#include "./Datenbank/mySQLModule.hpp"

//#define PGID PermutationGroup<S> 
#define PGID float

#define PG_SVW ValueWrapperByInitFunction<unsigned int, PGID>
#define PG_NRS NumberRangeSet<ValueWrapperByTemplateParameter<unsigned int, 0>, PG_SVW>
#define PG_CPS CartesianPowerSet<PG_NRS, PG_SVW>

//PermutationGroup: chain ring with addition and multiplication read from a file
template<typename S> class PermutationGroup;
template<typename S> class PermutationGroup_BaseSet;
template<typename S> struct PermutationGroup_Op;
//template<typename S> class Element<PermutationGroup_BaseSet<S> >;
//template<typename S> class AutomorphismGroup<PermutationGroup<S> >;
//template<typename S> class Element<AutomorphismGroup<PermutationGroup<S> > >;

//classes
template<typename S>
class PermutationGroup {
    public:
	typedef Semigroup_Atom<PermutationGroup_Op<S> > type;
};

template<typename S>
class PermutationGroup_BaseSet : public PG_CPS {
    public:
	static void init() {
	    PG_SVW::init(Cardinality<S>()());
	}
	static const unsigned int vector_mode = 0;
	static const unsigned int cycle_mode = 1;
	static unsigned int getDisplayMode() {
	    return displayMode;
	}
	static void setDiplayMode(unsigned int const dm) {
	    displayMode = dm;
	}
	typedef S PG_BaseSet; //BaseSet is already used in the parent class CartesianPowerSet
    private:
	static unsigned int displayMode; //0: vector, 1: cycle
};

template<typename S>
unsigned int PermutationGroup_BaseSet<S>::displayMode = 0;

template<typename S>
struct Base<PermutationGroup_BaseSet<S> > {
    typedef PG_CPS type;
};

template<typename S>
struct Base<PermutationGroup<S> > {
    typedef Semigroup<PermutationGroup_Op<S> > type;
};

template<typename S>
class Element<PermutationGroup_BaseSet<S> > : public Element<PG_CPS> {
    private:
	typedef PG_CPS Parent_;

    public:
	Element() : Element<Parent_>() { }
	Element(Element<Parent_> const & a) : Element<Parent_>(a) {}
};

template<typename S, typename Derived>
struct NameStream<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public NameStream<Root, Derived> {
    typedef Derived PGBS_;
    typedef typename PGBS_::PG_BaseSet S_;
    std::ostream & operator()(std::ostream & o) {
	return NameStream<S_>()(o << "S_");
    }
};

template<typename S, typename Derived>
class ElementName<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public ElementName<Root, Derived> {
    private:
	typedef Derived PGBS_;
    public:
    std::ostream & operator()(std::ostream & o, Element<PGBS_> const & v) const {
	switch(Derived::getDisplayMode()) {
	    case (Derived::vector_mode): {
		o << "[ ";
		size_t ub = PG_SVW::value();
		for(size_t i = 0; i < ub; i++) {
		    if (i != 0) {
			o << ", ";
		    }
		    o << UnrankElement<S>()(Coordinate<PGBS_>()(v,i).value());
		    //o << (Coordinate<PGBS_>()(v,i).value());
		}
		o << " ]";
					 }
		break;
	    case (Derived::cycle_mode):
		std::vector<bool> printed(PG_SVW::value(),false);
		unsigned int leftToPrint = PG_SVW::value();
		unsigned int nrCyclesPrinted = 0;
		unsigned int p;
		unsigned int begin = 0;
		while(leftToPrint > 0) {
		    for(;printed[begin];++begin);
		    if(Coordinate<Derived>()(v,begin).value() == begin) { //do not print 1-cycles
			printed[begin] = true;
			--leftToPrint;
			continue;
		    }
		    o << "(";
		    bool first = true;
		    p = begin;
		    do {
			if(!first) {
			    o << " ";
			} else {
			    first = false;
			}
			//o << p;
			o << UnrankElement<S>()(p);
			printed[p] = true;
			--leftToPrint;
			p = Coordinate<Derived>()(v,p).value();
		    } while(p != begin);
		    o << ")";
		    ++nrCyclesPrinted;
		}
		if(nrCyclesPrinted == 0) {
		    o << "id";
		}
	}
	return o;
    }
};

// set.hpp
//template<typename S, typename Derived>
//struct ElementName<PermutationGroup_BaseSet<S>, Derived> : public ElementName<Root, Derived> {
//    typedef Derived R_;
//    std::ostream & operator()(std::ostream & o, Element<R_> const & a) const {
//	return o << PermutationGroup<S>::getElementName(a.value());
//    }
//};

/*
template<typename S, typename Derived>
struct NameStream<PermutationGroup_BaseSet<S>, Derived> : public NameStream<Root, Derived> {
    std::ostream & operator()(std::ostream & o) const {
	return o << PermutationGroup<S>::getName();
    }
};*/

template<typename S, typename Derived>
struct Random<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public NameStream<Root, Derived> {
    typedef Derived PGBS_;
    Element<PGBS_> operator()() {
	Element<PGBS_> res;
	unsigned int ub = PG_SVW::value();
	for(unsigned int i = 0; i < ub; ++i) {
	    Coordinate<PGBS_>()(res,i) = i;
	}
	--ub;	
	for(unsigned int i = 0; i < ub; ++i) {
	    //std::swap<unsigned int>(Coordinate<PGBS_>()(res,i).value(),Coordinate<PGBS_>()(res,i + rand() % (ub - i + 1)).value());
	    std::swap<unsigned int>(res.data[i].value(),(res.data[i + rand() % (ub - i + 1)].value()));
	}
	return res;
    }
};

template<typename S, typename Derived>
struct Representative<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public NameStream<Root, Derived> {
    typedef Derived PGBS_;
    Element<PGBS_> operator()() {
	return NeutralElement<PermutationGroup<PGBS_> >()();
    }
};


template<typename S, typename Derived>
class Iterator<PermutationGroup_BaseSet<S>, Derived> :
    public boost::iterator_facade<
        Iterator<PermutationGroup_BaseSet<S>, Derived>,
        Element<Derived> const,
        boost::random_access_traversal_tag,
        Element<Derived> const &,
        signed int
    >
{
    friend class boost::iterator_core_access;
    friend struct Begin<PermutationGroup_BaseSet<S>, Derived>;
    friend struct End<PermutationGroup_BaseSet<S>, Derived>;
    friend struct IsAtBegin<PermutationGroup_BaseSet<S>, Derived>;
    friend struct IsAtEnd<PermutationGroup_BaseSet<S>, Derived>;
    public:
	Iterator(){} //uninitialized
        Iterator(unsigned int const n){
	    isAtEnd=false;
	    unsigned int c=Cardinality<S>()();
	    for (unsigned int i=0; i<c; i++){
		currElem.data[i].setValue(i);
	    }
	    unsigned int d=n;
	    while(d>0){
		increment();
		d--;
	    }
	};

    private:
	bool isAtEnd;
        Element<Derived> currElem;
        void increment(){
	    isAtEnd=!std::next_permutation(currElem.data.begin(),currElem.data.end(),Compare<PG_NRS>());
        }

        void decrement() {
	    std::prev_permutation(currElem.data.begin(),currElem.data.end(),Compare<PG_NRS>());
	    isAtEnd=false;
        }

        bool equal(Iterator<Derived> const & other) const {
            return Equals<PermutationGroup_BaseSet<S>,Derived>()(currElem,*other);
        }

        void advance(signed int const n){
	    int c=n;
	    while(c>0){
		increment();
		c--;
	    }
        }

        Element<Derived> const & dereference() const {
            return currElem;
        }
};

template<typename S, typename Derived>
struct Begin<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public Begin<Root, Derived> {
    public:
	Iterator<Derived> operator()() const {
	    return Iterator<Derived>(0);
	}
};

template<typename S, typename Derived>
struct End<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public Begin<Root, Derived> {
    public:
	Iterator<Derived> operator()() const {
	    Iterator<Derived> iter;
	    /*
	    unsigned int c=Cardinality<S>()();
	    unsigned int val=0;
	    while(c>0){
		c--;
		iter.currElem.data[c]=val;
		val++;
	    }
	    return iter;*/
	    iter.isAtEnd=true;
	    return iter;
	}
};

template<typename S, typename Derived>
struct IsAtBegin<PermutationGroup_BaseSet<S>, Derived, TypeConversionByImplementation> : public IsAtBegin<Root, Derived> {
    public:
	bool operator()(Iterator<Derived> const & i) const {
	    unsigned int c=Cardinality<S>()();
	    for (unsigned int j=0; j<c; j++){
		if (i.currElem.data[j].value()!=j){
		    return false;
		}
	    }
	    return true;
	}
};

template<typename S, typename Derived>
struct IsAtEnd<PermutationGroup_BaseSet<S> , Derived, TypeConversionByImplementation> : public IsAtEnd<Root, Derived> {
    public:
	bool operator()(Iterator<Derived> const & i) const {
	    return (i.isAtEnd);
	}
};

template<typename S, typename Derived>
struct Cardinality<PermutationGroup_BaseSet<S> , Derived, TypeConversionByImplementation> : public Cardinality<Root, Derived> {
    public:
	unsigned int operator()() const {
	    return mathHelper::factorial(PG_SVW::value());
	}
};


//additiveSemigroup
template<typename S>
struct PermutationGroup_Op : public Map<boost::mpl::vector<PermutationGroup_BaseSet<S>, PermutationGroup_BaseSet<S> >,PermutationGroup_BaseSet<S> > {
    typedef PermutationGroup_BaseSet<S> PGS_;
    typedef PermutationGroup<S> PG_;
    Element<PGS_> operator()(Element<PGS_> const & a, Element<PGS_> const & b) const {
	Element<PGS_> res;
	unsigned int ub = PG_SVW::value();
	for(unsigned int i = 0; i < ub; ++i) {
	    Coordinate<PGS_>()(res,i) = Coordinate<PGS_>()(a,Coordinate<PGS_>()(b,i).value()).value();
	}
	return res;
    }
};

template<typename S, typename Derived>
struct HasNeutralElement<Semigroup<PermutationGroup_Op<S> >, Derived> : public boost::mpl::bool_<true> {};

template<typename S, typename Derived>
struct HasInverseElements<Semigroup<PermutationGroup_Op<S> >, Derived> : public boost::mpl::bool_<true> {};

//template<typename S, typename Derived>
//struct IsCommutative<Semigroup<PermutationGroup_Op<S> >, Derived> : public boost::mpl::bool_<true> {};

template<typename S, typename Derived>
struct NeutralElement<Semigroup<PermutationGroup_Op<S> >, Derived> : public NeutralElement<Root, Derived> {
    typedef typename Derived::type::BaseSet PGS_;
    Element<PGS_> operator()() const {
	Element<PGS_> res;
	unsigned int ub = PG_SVW::value();
	for(unsigned int i = 0; i < ub; ++i) {
	    Coordinate<PGS_>()(res,i) = i;
	}
	return res;
    }
};

template<typename S, typename Derived>
struct EqualsNeutralElement<Semigroup<PermutationGroup_Op<S> >, Derived> : public EqualsNeutralElement<Root, Derived> {
    typedef typename Derived::type::BaseSet PGS_;
    bool operator()(Element<PGS_> const & a) const {
	unsigned int ub = PG_SVW::value();
	for(unsigned int i = 0; i < ub; ++i) {
	    if (Coordinate<PGS_>()(a,i).value() != i) {
		return false;
	    }
	}
	return true;
    }
};

template<typename S, typename Derived>
struct InverseElement<Semigroup<PermutationGroup_Op<S> >, Derived> : public InverseElement<Root, Derived> {
    typedef typename Derived::type::BaseSet PGS_;
    Element<PGS_> operator()(Element<PGS_> const & a) const {
	Element<PGS_> res;
	unsigned int ub = PG_SVW::value();
	for(unsigned int i = 0; i < ub; ++i) {
	    Coordinate<PGS_>()(res,Coordinate<PGS_>()(a,i).value()) = i;
	}
	return res;
    }
};

template<typename S, typename Derived>
struct Generators<Semigroup<PermutationGroup_Op<S> >, Derived> : public Generators<Root, Derived> {
    typedef typename Derived::type::BaseSet PGS_;
    std::vector<Element<PGS_> > operator()() const {
	std::vector<Element<PGS_> > res;
	unsigned int ub = PG_SVW::value();
	if (ub == 1) {
	    return res;
	}
	if (ub == 2) {
	    res.resize(1);
	} else {
	    res.resize(2);
	}
	Coordinate<PGS_>()(res[0],0) = 1;
	Coordinate<PGS_>()(res[0],1) = 0;
	if (ub == 2) return res;
	for (unsigned int i = 2; i < ub; ++i) {
	    Coordinate<PGS_>()(res[0],i) = i;
	}
	Coordinate<PGS_>()(res[1],0) = ub-1;
	for (unsigned int i = 1; i < ub; ++i) {
	    Coordinate<PGS_>()(res[1],i) = i-1;
	}
	return res;
    }
};

template<typename S>
struct Mul<PermutationGroup_BaseSet<S> > : public PermutationGroup_Op<S> {};

template<typename S>
struct Base<Semigroup<Mul<PermutationGroup_BaseSet<S> > > >{
    typedef PermutationGroup<S> type;
};

#undef PG_SVW
#undef PG_NRS
#undef PG_CPS

#endif
