/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef SEMIDIRECT_PRODUCT_HPP
#define SEMIDIRECT_PRODUCT_HPP

#include "cartesianProductSet.hpp"
#include "semigroup.hpp"
#include <boost/mpl/and.hpp>

template<typename NOp, typename HOp, typename Phi>
struct SemidirectProduct_BaseSet : public CartesianProductSet<typename NOp::ArgumentType1, typename HOp::ArgumentType1> {
    public:
	typedef typename NOp::ArgumentType1 N_BS_;
	typedef typename HOp::ArgumentType1 H_BS_;
	typedef NOp NOp_;
	typedef HOp HOp_;
	typedef Phi Phi_;
};

template<typename NOp, typename HOp, typename Phi>
struct Base<SemidirectProduct_BaseSet<NOp, HOp, Phi> > {
    typedef CartesianProductSet<typename NOp::ArgumentType1, typename HOp::ArgumentType1> type;
};

template<typename NOp, typename HOp, typename Phi>
class Element<SemidirectProduct_BaseSet<NOp, HOp, Phi> > : public Element<CartesianProductSet<typename NOp::ArgumentType1, typename HOp::ArgumentType1> > {
    public:
	typedef typename NOp::ArgumentType1 N_BS_;
	typedef typename HOp::ArgumentType1 H_BS_;
	typedef NOp NOp_;
	typedef HOp HOp_;
	typedef Phi Phi_;
    private:
	typedef CartesianProductSet<N_BS_, H_BS_> Parent_;

    public:
	Element() : Element<Parent_>() {}
	Element(Element<N_BS_> const & leftValueIn, Element<H_BS_> const & rightValueIn) : Element<Parent_>(leftValueIn, rightValueIn) {};
	Element(std::pair<Element<N_BS_>,Element<H_BS_> > const & pairIn) : Element<Parent_>(pairIn) {};
};

template<typename NOp, typename HOp, typename Phi, typename Derived>
class NameStream<SemidirectProduct_BaseSet<NOp, HOp, Phi>, Derived> : public NameStream<Root, Derived> {
    private:
	typedef Derived SDP_;
	typedef typename SDP_::N_BS_ N_BS_;
	typedef typename SDP_::H_BS_ H_BS_;
    public:
        std::ostream & operator()(std::ostream & o) {
            return NameStream<H_BS_>()(NameStream<N_BS_>()(o) << " ><| ");
        }
};

template<typename NOp, typename HOp, typename Phi>
class SemidirectProduct_Op :
    public Map<
        boost::mpl::vector<
	    SemidirectProduct_BaseSet<NOp, HOp, Phi>,
	    SemidirectProduct_BaseSet<NOp, HOp, Phi>
	>,
	SemidirectProduct_BaseSet<NOp, HOp, Phi>
    >
{
    public:
	typedef NOp NOperator;
	typedef HOp HOperator;
	typedef Phi DefiningHomomorphism;
    private:
	typedef typename NOp::ArgumentType1 N_BS_;
	typedef typename HOp::ArgumentType1 H_BS_;
	typedef SemidirectProduct_BaseSet<NOp, HOp, Phi> SDP_BS_;
    public:
	Element<SDP_BS_> operator()(Element<SDP_BS_> const & a, Element<SDP_BS_> const & b) const {
	    return Element<SDP_BS_>(NOp()(a.left(),Phi()(a.right(),b.left())),HOp()(a.right(),b.right()));
	}
};

template<typename NOp, typename HOp, typename Phi>
struct SemidirectProduct {
    BOOST_MPL_ASSERT_RELATION((Phi::NumberOfArguments::value), ==, 2);
    BOOST_MPL_ASSERT((boost::is_same<typename Phi::ArgumentType1, typename HOp::ArgumentType1>));
    BOOST_MPL_ASSERT((boost::is_same<typename Phi::ArgumentType2, typename NOp::ArgumentType1>));
    BOOST_MPL_ASSERT((boost::is_same<typename Phi::ResultType, typename NOp::ArgumentType1>));
    typedef Semigroup_Atom<SemidirectProduct_Op<NOp,HOp,Phi> > type;
    typedef typename NOp::ArgumentType1 N;
    typedef typename HOp::ArgumentType1 H;
};

template<typename NOp, typename HOp, typename Phi>
struct Base<SemidirectProduct<NOp, HOp, Phi> > {
    typedef Semigroup<SemidirectProduct_Op<NOp,HOp,Phi> > type;
};

template<typename NOp, typename HOp, typename Phi, typename Derived>
struct HasNeutralElement<Semigroup<SemidirectProduct_Op<NOp,HOp,Phi> >, Derived> :
boost::mpl::and_<HasNeutralElement<Semigroup<typename Derived::type::Operator::NOperator> >,HasNeutralElement<Semigroup<typename Derived::type::Operator::HOperator> > >::type {};

template<typename NOp, typename HOp, typename Phi, typename Derived>
struct HasInverseElements<Semigroup<SemidirectProduct_Op<NOp,HOp,Phi> >, Derived> :
boost::mpl::and_<HasInverseElements<Semigroup<typename Derived::type::Operator::NOperator> >,HasInverseElements<Semigroup<typename Derived::type::Operator::HOperator> > >::type {};

template<typename NOp, typename HOp, typename Phi, typename Derived>
struct NeutralElement<Semigroup<SemidirectProduct_Op<NOp,HOp,Phi> >, Derived> {
        typedef Derived SDP_;
	typedef typename SDP_::type::BaseSet SDP_BS_;
	typedef typename SDP_BS_::LeftBaseSet N_BS_;
	typedef typename SDP_BS_::RightBaseSet H_BS_;
	typedef typename SDP_::type::Operator SDP_Op_;
	typedef typename SDP_Op_::NOperator NOp_;
	typedef typename SDP_Op_::HOperator HOp_;
	BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<NOp_> >));
	BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<HOp_> >));
	Element<SDP_BS_> operator()() const {
	    return Element<SDP_BS_>(NeutralElement<Semigroup<NOp_> >()(),NeutralElement<Semigroup<HOp_> >()());
	}
};

template<typename NOp, typename HOp, typename Phi, typename Derived>
struct EqualsNeutralElement<Semigroup<SemidirectProduct_Op<NOp,HOp,Phi> >, Derived> {
        typedef Derived SDP_;
	typedef typename SDP_::type::BaseSet SDP_BS_;
	typedef typename SDP_BS_::LeftBaseSet N_BS_;
	typedef typename SDP_BS_::RightBaseSet H_BS_;
	typedef typename SDP_::type::Operator SDP_Op_;
	typedef typename SDP_Op_::NOperator NOp_;
	typedef typename SDP_Op_::HOperator HOp_;
	BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<NOp_> >));
	BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<HOp_> >));
	bool operator()(Element<SDP_BS_> const & a) const {
	    return (EqualsNeutralElement<Semigroup<NOp_> >()(a.left()) && EqualsNeutralElement<Semigroup<HOp_> >()(a.right()));
	}
};

template<typename NOp, typename HOp, typename Phi, typename Derived>
struct InverseElement<Semigroup<SemidirectProduct_Op<NOp,HOp,Phi> >, Derived> {
        typedef Derived SDP_;
	typedef typename SDP_::type::BaseSet SDP_BS_;
	typedef typename SDP_BS_::LeftBaseSet N_BS_;
	typedef typename SDP_BS_::RightBaseSet H_BS_;
	typedef typename SDP_::type::Operator SDP_Op_;
	typedef typename SDP_Op_::NOperator NOp_;
	typedef typename SDP_Op_::HOperator HOp_;
	BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<NOp_> >));
	BOOST_MPL_ASSERT((HasInverseElements<Semigroup<NOp_> >));
	BOOST_MPL_ASSERT((HasNeutralElement<Semigroup<HOp_> >));
	BOOST_MPL_ASSERT((HasInverseElements<Semigroup<HOp_> >));
	Element<SDP_BS_> operator()(Element<SDP_BS_> const & a) const {
	    Element<H_BS_> hInv = InverseElement<Semigroup<HOp_> >()(a.right());
	    return Element<SDP_BS_>(Phi()(hInv, InverseElement<Semigroup<NOp_> >()(a.left())),hInv);
	}
};

template<typename NOp, typename HOp, typename Phi, typename Derived>
struct Mul<SemidirectProduct_BaseSet<NOp, HOp, Phi>, Derived> : public SemidirectProduct_Op<typename Derived::NOp_, typename Derived::HOp_, typename Derived::Phi_> {};

#endif
