/******************************************************************
Copyright 2010 by Michael Kiermaier and Johannes Zwanzger, Bayreuth
******************************************************************/

#ifndef _SEMIGROUP_BY_GENERATORS_HPP
#define _SEMIGROUP_BY_GENERATORS_HPP

#include "set.hpp"
#include "semigroup.hpp"
#include "action.hpp"
#include "additiveSemigroup.hpp"
#include "multiplicativeSemigroup.hpp"
#include <vector>
#include <boost/mpl/assert.hpp>
#include <iostream>
#include <boost/iterator/iterator_facade.hpp>
#include <queue>

template<typename SG, typename id> class SemigroupByGenerators_BaseSet : public SG::type::BaseSet {
    public: 
	typedef SG BaseSemigroup;
	//static std::vector<Element<typename SG::type::BaseSet> > generators;
	static std::vector<Element<SemigroupByGenerators_BaseSet<SG, id> > > generators;
	static void init(std::vector<Element<typename SG::type::BaseSet> > const & generators_in) {
	    typedef typename std::vector<Element<typename SG::type::BaseSet> >::const_iterator ItC_;
	    typedef typename std::vector<Element<SemigroupByGenerators_BaseSet<SG, id> > >::iterator It_;
	    generators.resize(generators_in.size());
	    ItC_ ub = generators_in.end();
	    It_ gen_it = generators.begin();
	    for(ItC_ it = generators_in.begin(); it != ub; ++it) {
		(*gen_it) = (*it);
		++gen_it;
	    }
	}
};

template<typename SG, typename id>
//std::vector<Element<typename SG::type::BaseSet> > SemigroupByGenerators_BaseSet<SG, id>::generators;
std::vector<Element<SemigroupByGenerators_BaseSet<SG, id> > > SemigroupByGenerators_BaseSet<SG, id>::generators;

//root.hpp
template<typename SG, typename id> 
struct Base<SemigroupByGenerators_BaseSet<SG, id> > {
    typedef typename SG::type::BaseSet type;
};

/*
template<typename Op>
struct TypeConversion<SemigroupByGenerators<Op>, Equals<Root> > {
    typedef TypeConversionByCast type;
};
*/

//set.hpp
template<typename SG, typename id>
class Element<SemigroupByGenerators_BaseSet<SG, id> > : public Element<typename SG::type::BaseSet> {
    public:
	Element(){}
	Element(Element<typename SG::type::BaseSet> const & a) : Element<typename SG::type::BaseSet>(a) {}
    
};

template<typename SG, typename id, typename Derived>
struct NameStream<SemigroupByGenerators_BaseSet<SG,id>, Derived> : public NameStream<Root, Derived> {
    typedef Derived SGBG_;
    typedef typename SGBG_::BaseSemigroup SG_;
    typedef typename SG_::type::BaseSet S_;
    std::ostream & operator()(std::ostream & o) const {
	o << "Subsemigroup of " << Name<S_>()() << " generated by\n";
	typedef typename std::vector<Element<SGBG_> >::const_iterator Iter;
	for(Iter i = SGBG_::generators.begin(); i != SGBG_::generators.end(); ++i) {
	    o << (*i) << "\n";
	}
	return o;
    }
};

template<typename SG, typename id, typename Derived>
class Iterator<SemigroupByGenerators_BaseSet<SG, id>, Derived> : 
public boost::iterator_facade<
    Iterator<SemigroupByGenerators_BaseSet<SG, id>, Derived>,
    Element<Derived> const,
    boost::incrementable_traversal_tag
>
{
    friend class boost::iterator_core_access;
    friend class Begin<SemigroupByGenerators_BaseSet<SG, id>, Derived>;
    friend class IsAtEnd<SemigroupByGenerators_BaseSet<SG, id>, Derived>;

    private:
	typedef Derived PGBG_BS;
	typedef typename Derived::BaseSemigroup SG_;
	typedef typename SG_::type::BaseSet S_;

    public:
	Iterator(){};

    private:
	HASH_SET(Element<S_>) visited;
	std::queue<Element<S_> > newElements;

    public:
	void increment() {
	    Element<S_> a = newElements.front();
	    newElements.pop();
	    typedef typename std::vector<Element<S_> >::const_iterator iter;
	    for(iter i = PGBG_BS::generators.begin(); i != PGBG_BS::generators.end(); ++i) {
		Element<S_> b = typename SG_::type::Operator()(*i,a);
		if (visited.count(b) == 0) {
		    newElements.push(b);
		    visited.insert(b);
		}
	    }
	}

	bool equal(Iterator<SemigroupByGenerators_BaseSet<SG_, id> > const & other) const {
	    return (newElements.empty() && other.newElements.empty()) || (newElements.front() == other.newElements.front());
	}

	Element<Derived> const & dereference() const {
	    return (Element<PGBG_BS> const &)(newElements.front());
	}
};

template<typename SG, typename id, typename Derived>
class Begin<SemigroupByGenerators_BaseSet<SG, id>, Derived> : public Begin<Root, Derived> {
    private:
	typedef Derived PGBG_BS;
	typedef typename Derived::BaseSemigroup SG_;
	typedef typename SG_::type::BaseSet S_;

    public:
	Iterator<PGBG_BS> operator()() const {
	    Iterator<PGBG_BS> it;
	    typedef typename std::vector<Element<S_> >::const_iterator iter;
	    for(iter i = PGBG_BS::generators.begin(); i != PGBG_BS::generators.end(); ++i) {
		it.newElements.push(*i);
		it.visited.insert(*i); //??
	    }
	    return it;
	}
};

template<typename SG, typename id, typename Derived>
class IsAtEnd<SemigroupByGenerators_BaseSet<SG, id>, Derived> : public IsAtEnd<Root, Derived> {
    private:
	typedef Derived PGBG_BS;
	typedef typename Derived::BaseSemigroup SG_;
	typedef typename SG_::type::BaseSet S_;

    public:
	bool operator()(Iterator<PGBG_BS> const & it) const {
	    return it.newElements.empty();
	}
};

template<typename SG, typename id, typename Derived>
class Random<SemigroupByGenerators_BaseSet<SG, id>, Derived> : public Random<Root, Derived> {
    private:
	typedef Derived PGBG_BS;
	typedef typename Derived::BaseSemigroup SG_;
	typedef typename SG_::type::BaseSet S_;

    public:
	Element<PGBG_BS> operator()() const {
	    Element<PGBG_BS> res;
	    if (PGBG_BS::generators.empty()) {
		res = NeutralElement<SG_>()();
	    } else {
		res = PGBG_BS::generators[rand() % PGBG_BS::generators.size()];
		unsigned int ub = rand() % 100;
		for(unsigned int i = 0; i < ub; ++i) {
		    InRightPlace<SG_>()(Power<SG_>()(PGBG_BS::generators[rand() % PGBG_BS::generators.size()],rand()),res);
		}
	    }
	    return res;
	}
};

//semigroup.hpp
template<typename SG, typename id>
struct SemigroupByGenerators_Operator : public Map<boost::mpl::vector<SemigroupByGenerators_BaseSet<SG, id>, SemigroupByGenerators_BaseSet<SG, id> >, SemigroupByGenerators_BaseSet<SG,id> > {
    Element<SemigroupByGenerators_BaseSet<SG, id> > operator()(Element<SemigroupByGenerators_BaseSet<SG, id> > const & a, Element<SemigroupByGenerators_BaseSet<SG, id> > const & b) const {
	return Element<SemigroupByGenerators_BaseSet<SG, id> >(typename SG::type::Operator()(Element<typename SG::type::BaseSet>(a),Element<typename SG::type::BaseSet>(b)));
    }
};

template<typename SG, typename id>
struct SemigroupByGenerators : public SG {
    typedef Semigroup_Atom<SemigroupByGenerators_Operator<SG, id> > type;
};

template<typename SG, typename id, typename Derived>
struct Generators<Semigroup<SemigroupByGenerators_Operator<SG, id> >, Derived> {
    typedef Derived SGBG_;
    typedef typename SGBG_::type::BaseSet SGBG_BS_;
    std::vector<Element<SGBG_BS_> > operator()() const {
	return SGBG_BS_::generators;
    }
};

template<typename SG, typename id> 
struct Base<SemigroupByGenerators<SG, id> > {
    typedef Semigroup<SemigroupByGenerators_Operator<SG, id> > type;
};

template<typename SG, typename id> 
struct Base<Semigroup<SemigroupByGenerators_Operator<SG, id> > > {
    typedef SG type;
};

template<typename SG, typename id> 
struct Base<SemigroupByGenerators_BaseSet<SG, id>, Cardinality<Root> > {
    typedef Root type;
};

template<typename SG, typename id> 
struct Base<SemigroupByGenerators_BaseSet<SG, id>, RankElement<Root> > {
    typedef Root type;
};

template<typename SG, typename id> 
struct Base<SemigroupByGenerators_BaseSet<SG, id>, UnrankElement<Root> > {
    typedef Root type;
};

template<typename S, typename id>
struct Add<SemigroupByGenerators_BaseSet<Semigroup<Mul<S> >, id> > : public SemigroupByGenerators_Operator<Semigroup<Add<S> >, id> {};

template<typename S, typename id>
struct Mul<SemigroupByGenerators_BaseSet<Semigroup<Mul<S> >, id> > : public SemigroupByGenerators_Operator<Semigroup<Mul<S> >, id> {};

#endif
