/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef ABSTRACT_EVALUATION_TABLE_HPP
#define ABSTRACT_EVALUATION_TABLE_HPP

#include <math.h>
#include <string>
#include <sstream>
#include <iomanip>

#include "defs.h"
#include "homogenousWeightTable.hpp"
#include "twoDimArray.hpp"
#include "messageType.hpp"
#include "polynom.h"
#include "abstractWeightTable.hpp"

template<class CHAINRING>
class AbstractEvaluationTable{
    public:
    AbstractEvaluationTable();
    virtual void initialize(MessageType<CHAINRING>&, AbstractWeightTable<CHAINRING>*, unsigned int, WEIGHT_SUM)=0;
    EVAL_TYPE get_eval(unsigned int, WEIGHT_SUM);
    std::string getAsString();

    protected:
        bool initialized;
        unsigned int n;
        WEIGHT_SUM d;
        TwoDimArray<EVAL_TYPE> eval; 
};

template<class CHAINRING>
AbstractEvaluationTable<CHAINRING>::AbstractEvaluationTable(){
    this->initialized=false;
}

template<class CHAINRING>
EVAL_TYPE AbstractEvaluationTable<CHAINRING>::get_eval(unsigned int n, WEIGHT_SUM w){
    return this->eval[n][w];
}

template<class CHAINRING>
std::string AbstractEvaluationTable<CHAINRING>::getAsString(){
    std::stringstream s;
    if (!this->initialized){
        s<<"Error: AbstractEvaluationTable not initialized!"<<std::endl;
    }
    else{
	s<<std::setprecision(4);
	s<<"     ";
        for (WEIGHT_TYPE j=0; j<=this->d; j++){
	    s<<std::setw(6)<<(int)j<<" ";
	}
	s<<std::endl<<std::endl;
        for (unsigned int i=0; i<=this->n; i++){
	    s<<std::setw(3)<<i<<": ";
            for (WEIGHT_TYPE j=0; j<=this->d; j++){
		s<<std::setw(6)<<exp((double)this->eval[i][j]/EVAL_SHIFT)<<" ";
            }
            s<<std::endl;
        }        
    }
    return s.str();
}

#endif

