/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "arcEvaluationTable.h"
#include "helpFunctions.h"
#include <algorithm>
#include <sstream>
#include <math.h>
#include <iostream>
#include <iomanip>

ArcEvaluationTable::ArcEvaluationTable(){
    this->initialized=false;
}

ArcEvaluationTable::~ArcEvaluationTable(){
}

void ArcEvaluationTable::initialize(int n, int v, int e1, int e2, int delta, int d1, int d){
    this->initialized=true; 
    this->n=n;
    this->d=d;
    table.init(this->n+1,this->d+1);
    for (int _n=0; _n<=this->n; _n++){ //the number of columns which shall be added
	if (_n<d1){
	    for (int d2=0; d2<=d; d2++){
		this->table[_n][d2]=MINUS_INFTY_LL;
	    }
	    continue;
	}
	double* tempTable=new double[d+1];
	double b=calcDistNumber(d1,e1)*calcDistNumber(_n-d1,v-e1);
	for (int d2=0; d2<=d; d2++){
	    tempTable[d2]=0.0;
	}
	for (int d2=0; d2<=_n; d2++){ //the sum to obtain in the inequation row
	    double a=0.0; 
	    for (int j=(std::max)(0,d1+d2-_n); j<=(std::min)(d1,d2); j++){
		a+=calcDistNumber(j,delta)*calcDistNumber(d1-j,e1-delta)*calcDistNumber(d2-j,e2-delta)*calcDistNumber(_n-d1-(d2-j),v-e1-(e2-delta));
	    }
	    tempTable[std::min(d2,this->d)]+=(a/b);
	}
	int d2=this->d-1;
	while(true){
	    tempTable[d2]+=tempTable[d2+1];
	    if (d2==0){
		break;
	    }
	    d2--;
	}
	for (int d2=0; d2<=d; d2++){
	    if (tempTable[d2]==0.0){
		this->table[_n][d2]=MINUS_INFTY_LL;
	    }
	    else{
		this->table[_n][d2]=(EVAL_TYPE)(log(tempTable[d2])*EVAL_SHIFT);
	    }
	}
    }
}

EVAL_TYPE ArcEvaluationTable::get_eval(int _n, int d2){
    return this->table[_n][d2];
}

std::string ArcEvaluationTable::getTableAsString(){
    std::stringstream ss;
    if (!this->initialized){
        ss<<"Error: AbstractEvaluationTable not initialized!"<<std::endl;
    }
    else{
	ss<<std::setprecision(4);
	ss<<"     ";
        for (WEIGHT_SUM d2=0; d2<=(WEIGHT_SUM)this->d; d2++){
	    ss<<std::setw(6)<<(int)d2<<" ";
	}
	ss<<std::endl<<std::endl;
        for (int _n=0; _n<=this->n; _n++){
	    ss<<std::setw(3)<<_n<<": ";
    	    for (int d2=0; d2<this->d; d2++){
		ss<<std::setw(6)<<exp(double(this->table[_n][d2])/EVAL_SHIFT)<<" ";
	    }
	    ss<<std::setw(6)<<exp(double(this->table[_n][this->d])/EVAL_SHIFT)<<std::endl;
	}
    }
    return ss.str();
}

