/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef CODEWORD_MANAGER_HPP
#define CODEWORD_MANAGER_HPP

#include "abstractCodewordManager.hpp"
#include "stochasticEvaluationTable.hpp"
#include "solutionDensityTable.hpp"
#include "defs.h"

#include <string>
#include <algorithm>
#include <sstream>
#include <vector>

template <class CHAINRING>
class CodewordManager : public AbstractCodewordManager<CHAINRING>{
    public:
        void initializeEval();
	EVAL_TYPE getEval(COLUMN_NO);
	std::pair<EVAL_TYPE,HASH_TYPE> getEvalAndHash(COLUMN_NO);
    private:
	std::vector<AbstractEvaluationTable<CHAINRING>* > evalTables; 	
};

//******************** implementation ********************
template <class CHAINRING>
EVAL_TYPE CodewordManager<CHAINRING>::getEval(COLUMN_NO c){
    EVAL_TYPE eval=0;    
    unsigned int current_n=this->columnHistory.size()+1;
    unsigned int n_diff=this->n-current_n;
    WEIGHT_TYPE* incidenceRow=this->incidence_matrix[c];
    for (int i=this->type_sections.size()-1; i>=0; i--){
        AbstractEvaluationTable<CHAINRING>* evalTable=this->evalTables[i];
	NumberListEntry* p_end=this->type_sections[i]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
	    int temp=this->codewordWeight[p->nr]+incidenceRow[p->nr];
	    int w_diff=(std::max)(0,(int)this->d-temp);
	    EVAL_TYPE ev=evalTable->get_eval(n_diff,w_diff);
	    if (ev==MINUS_INFTY_LL){
		//p->sortToFront(); //->AbstractCWM::pop_column must call undoRemoveToEnd();
		return MINUS_INFTY_LL;
	    }
	    eval+=ev;
        }
    }
    return eval;
}

template <class CHAINRING>
std::pair<EVAL_TYPE,HASH_TYPE> CodewordManager<CHAINRING>::getEvalAndHash(COLUMN_NO c){
    EVAL_TYPE eval=0;    
    unsigned int current_n=this->columnHistory.size()+1;
    unsigned int n_diff=this->n-current_n;
    WEIGHT_TYPE* incidenceRow=this->incidence_matrix[c];
    this->topNHC->resetHash();
    for (int i=this->type_sections.size()-1; i>=0; i--){
        AbstractEvaluationTable<CHAINRING>* evalTable=this->evalTables[i];
	NumberListEntry* p_end=this->type_sections[i]->messages;
	for (NumberListEntry* p=p_end->nextEntry; p!=p_end; p=p->nextEntry){
	    int temp=this->codewordWeight[p->nr]+incidenceRow[p->nr];
	    int w_diff=(std::max)(0,(int)this->d-temp);
	    EVAL_TYPE ev=evalTable->get_eval(n_diff,w_diff);
	    this->nbcByMessageNumber[p->nr]->hash+=(HASH_TYPE)(ev);
	    if (ev==MINUS_INFTY_LL){
		//p->sortToFront(); //->AbstractCWM::pop_column must call undoRemoveToEnd();
		return std::pair<EVAL_TYPE,HASH_TYPE>(MINUS_INFTY_LL,0);
	    }
	    eval+=ev;
        }
    }
    return std::pair<EVAL_TYPE,HASH_TYPE>(eval,this->topNHC->getHash());
}

//initializeSearch must have been called before in order to set n and d!
template <class CHAINRING>
void CodewordManager<CHAINRING>::initializeEval(){
    for (unsigned int i=0; i<this->evalTables.size(); i++){
	delete this->evalTables[i];
    }
    this->evalTables.clear();
    for (unsigned int i=0; i<this->type_sections.size(); i++){	
	//consider the weight distribution of the already chosen columns
	WEIGHT_SUM minWeight=this->d;
	NumberListEntry* p_end=this->type_sections[i]->messages;
	NumberListEntry* p=p_end->nextEntry;
	if (p!=p_end){
	    minWeight=this->codewordWeight[p->nr];
	    for (p=p->nextEntry; p!=p_end; p=p->nextEntry){
    		minWeight=(std::min)(minWeight,this->codewordWeight[p->nr]);
	    }
	}
	AbstractEvaluationTable<CHAINRING>* s=(AbstractEvaluationTable<CHAINRING>*)(new SolutionDensityTable<CHAINRING>());
	if (minWeight<this->d){
    	    s->initialize(this->type_sections[i]->messageType,this->weightTable,this->n-this->columnHistory.size(),this->d-minWeight);
	}
	else{
    	    s->initialize(this->type_sections[i]->messageType,this->weightTable,this->n-this->columnHistory.size(),0);
	}
	this->evalTables.push_back(s);
    }
}

#endif

