/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef GRAY_LIKE_ALGORITHM_HPP
#define GRAY_LIKE_ALGORITHM_HPP

#include <vector>

template<class CONTAINER> class PingPongIterator;
template<class CONTAINER> class grayLikeAlgorithm;

template<class CONTAINER>
class PingPongIterator{
public:
    PingPongIterator(CONTAINER*);
    typename CONTAINER::value_type resetToBegin();
    bool next(typename CONTAINER::value_type&);

private:
    bool forwardDirection;
    CONTAINER* container;
    typename CONTAINER::iterator iterF;
    typename CONTAINER::iterator endIterF;
    typename CONTAINER::reverse_iterator iterB;
    typename CONTAINER::reverse_iterator endIterB;
};

template<class CONTAINER>
PingPongIterator<CONTAINER>::PingPongIterator(CONTAINER* c){
    this->container=c;
    this->endIterF=this->container->end();
    this->endIterB=this->container->rend();
}

template<class CONTAINER>
typename CONTAINER::value_type PingPongIterator<CONTAINER>::resetToBegin(){
    this->forwardDirection=true;
    this->iterF=this->container->begin();
    typename CONTAINER::value_type res=*this->iterF;
    this->iterF++;
    return res;
}

template<class CONTAINER>
bool PingPongIterator<CONTAINER>::next(typename CONTAINER::value_type& e){
    if(this->forwardDirection){
	if (this->iterF==this->endIterF){
	    this->forwardDirection=false;
	    this->iterB=this->container->rbegin();
	    e=*this->iterB;
	    this->iterB++;
	    return false;
	}
	e=*this->iterF;
	this->iterF++;
	return true;
    }
    else{
	if (this->iterB==this->endIterB){
	    this->forwardDirection=true;
	    this->iterF=this->container->begin();
	    e=*this->iterF;
	    this->iterF++;
	    return false;
	}
	e=*this->iterB;
	this->iterB++;
	return true;
    }
}

template<class CONTAINER>
class GrayLikeAlgorithm{
public:
    void initialize(std::vector<CONTAINER*>); 
    std::vector<typename CONTAINER::value_type> resetToBegin();
    int next(std::vector<typename CONTAINER::value_type>&);

private:
    std::vector<PingPongIterator<CONTAINER> > ppIters;
    std::vector<typename CONTAINER::value_type> result;
}; 

template<class CONTAINER>
void GrayLikeAlgorithm<CONTAINER>::initialize(std::vector<CONTAINER*> c){
    this->ppIters.clear();
    typename std::vector<CONTAINER*>::iterator iter;
    typename std::vector<CONTAINER*>::iterator endIter;
    iter=c.begin();
    endIter=c.end();
    while(iter!=endIter){
	PingPongIterator<CONTAINER> ppi(*iter);
	this->ppIters.push_back(ppi);
	iter++;
    }
}

template<class CONTAINER>
std::vector<typename CONTAINER::value_type> GrayLikeAlgorithm<CONTAINER>::resetToBegin(){
    int size=this->ppIters.size();
    this->result.clear();
    for (int i=0; i<size; i++){
	this->result.push_back(this->ppIters[i].resetToBegin());
    }
    return this->result;
}

template<class CONTAINER>
int GrayLikeAlgorithm<CONTAINER>::next(std::vector<typename CONTAINER::value_type>& v){
    int i;
    int size=this->ppIters.size();
    typename std::vector<typename CONTAINER::value_type>::iterator iter_res=this->result.begin();
    for (i=0; i<size; i++){
	if (this->ppIters[i].next(*iter_res)){
	    break;
	}
	iter_res++;
    }
    v=this->result;
    return (size-i);
}

#endif

