/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "helpFunctions.h"

#ifdef LINUX
    #include <sys/time.h>
    #include <stdio.h>
    #include <termios.h>
    #include <unistd.h>
#else 
    #include <winsock2.h>
    #include <sys/timeb.h>
    #include <conio.h>
#endif

double faculty(int n){
    double res=1.0;
    for (int i=2; i<=n; i++){
	res*=(double)(i);
    }
    return res;
}

unsigned int myPowerFunc(unsigned int b, unsigned int e){
    unsigned int res=1;
    while(e!=0){
        res*=b;
	e--;
    }
    return res;
}

double calcDistNumber(int k, int z){
    if (z<0 || k<0){
	return 0.0;
    }
    if (k==0){
	return 1.0;
    }
    if (z==0){
	return 0.0;
    }
    double res=1.0;
    for (int i=1; i<=k; i++){
	res*=(double)((z-1)+i);
	res/=(double(i));
    }
    return res;
}

unsigned int logarithm(unsigned int n, unsigned int base){
    unsigned int res=0;
    unsigned int pow=1;
    while(pow<n){
	pow*=base;
	res++;
    }
    return res;
}

//Ersetzt einen String
std::string ReplaceString(const std::string &stringSearchString, const std::string &stringReplaceString, std::string stringStringToReplace){   
    std::string::size_type pos = stringStringToReplace.find(stringSearchString, 0);
    int intLengthSearch = stringSearchString.length();
    while(std::string::npos!=pos)
    {
	stringStringToReplace.replace(pos, intLengthSearch, stringReplaceString);
	pos = stringStringToReplace.find(stringSearchString, 0);

    }
    return stringStringToReplace;
}

unsigned int countOccurence(const std::string &searchString, const std::string& inStr){
    unsigned int counter=0;
    std::string::size_type pos = inStr.find(searchString, 0);
    while(std::string::npos!=pos)
    {
	counter++;
	pos = inStr.find(searchString, pos+1);
    }
    return counter;
}

unsigned int twistUint(unsigned int u){
    const unsigned long long mod=4294967291; //thats the biggest prime below 2^32
    const unsigned long long lin=2147483647; //2^31-1; thats a prime, too, but this should not matter
    if (u>=mod){
	return u;
    }
    unsigned long long u_as_ll=(unsigned long long)(u);
    return (unsigned int)(((u_as_ll*u_as_ll)+u_as_ll*lin)%mod);
}

unsigned long long hashTwist(unsigned long long i){
    unsigned int l=(unsigned int)(i&0x00000000FFFFFFFF);
    unsigned int h=(unsigned int)((i&0xFFFFFFFF00000000)>>32);
    return (((unsigned long long)(twistUint(h)))<<32)+twistUint(l);    
}

void waitSomeTime(unsigned int nanosecs){
    #ifdef LINUX
	timespec t;
	t.tv_sec=0; //seconds
	t.tv_nsec=nanosecs;
	nanosleep(&t,NULL);
    #else //windows version
	unsigned int ms=(std::max)(1,(int)(nanosecs/1000000));
	::Sleep(ms);
    #endif
}

std::string getNextWord(const std::string& s, size_t& pos){
    std::string res;
    size_t pos2=s.find(" ",pos);
    if (pos2!=std::string::npos){
	res=s.substr(pos,pos2-pos);
	pos=pos2+1;
    }
    else{
	res=s.substr(pos);
	pos=pos2;
    }
    return res;
}

std::vector<std::string> breakDownToWords(const std::string& s){
    size_t pos=0;
    std::vector<std::string> res;
    while(pos!=std::string::npos){
	res.push_back(getNextWord(s,pos));
    }
    return res;
}

unsigned int sdbmHash(unsigned char *str, unsigned int length){
    unsigned int hash = 0;
    int c;
    while (length>0){
	c = *str++;
	hash = c + (hash << 6) + (hash << 16) - hash;
	length--;
    }
    return hash;
}

char readCharacter(){
#ifdef LINUX
    struct termios oldt,
	newt;
    char            ch;
    tcgetattr( STDIN_FILENO, &oldt );
    newt = oldt;
    newt.c_lflag &= ~( ICANON | ECHO );
    tcsetattr( STDIN_FILENO, TCSANOW, &newt );
    ch = getchar();
    tcsetattr( STDIN_FILENO, TCSANOW, &oldt );
    return ch;
#else
    return _getch();
#endif
}
