/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "heurico2.h"
#include "helpFunctions.h"
#include "searchAlgorithm.hpp"
#include "codewordManager.hpp"
#include "arcSearchManager.hpp"
#include "log.h"
#include "defs.h"
#include "abstractPreselectionModule.h"
#include "defaultPreselectionModule.h"
#include "setwisePreselectionModule.h"
#include "./Ringe/tableChainRing.hpp"
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

const double version=1.16;
std::stringstream callingData;

int main(int argc, char* argv[]){
    typedef TableChainRing<boost::mpl::int_<0> > CHAINRING;
    callingData<<std::setprecision(3)<<std::showpoint;
    callingData<<"%version "<<version<<std::endl;
    callingData<<"%arguments: ";
    for (int i=0; i<argc; i++){
	callingData<<argv[i]<<"; ";
    }
    callingData<<std::endl;
    callingData<<"%compiler switches: ";
    callingData<<std::endl;
    std::string ringFileName("");
    std::string solFileName("");
    std::string pcFileName("");
    std::string selFileName("");
    std::string iMatFileName("");
    int n=-1;
    bool autoSearchCodes=false;
    bool autoSearchArcs=false;
    bool additiveOnly=false;
    int autosearchcodes_upper_n=MAX_LENGTH;
    WEIGHT_SUM d=0;
    unsigned int maxHashDepth=MAX_LENGTH;
    int k=-1;
    std::vector<PERIOD> lambda;
    PERIOD minPeriod_cols=1;
    PERIOD minPeriod_mess=1;
    AbstractCodewordManager<CHAINRING>* cwm;
    unsigned int whichWeight=WEIGHT_UNDEFINED;
    unsigned int maxBranch=0xFFFFFFFF;
    double alterBound=DEFAULT_ALTER_BOUND;
    double timeout=DEFAULT_TIMEOUT;
    unsigned int evalHashSize=DEFAULT_EVALHASH_SIZE;
    unsigned int columnHashSize=DEFAULT_COLUMNHASH_SIZE;
    unsigned int outputLevel=OUTPUT_NORMAL;
    bool htmlTree=false;
    bool arcs=false;
    bool error=false;
    DefaultPreselectionModule dpsm;
    SetwisePreselectionModule spsm;
    AbstractPreselectionModule* psm=&dpsm;

    std::string learnFileName;
    for (int i=1; i<argc; i++){
	if (std::string(argv[i])=="ring" || std::string(argv[i])=="-ring"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no filename for the ring given!\n"));
		error=true;
	    }
	    else{
    		ringFileName=std::string(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="sol" || std::string(argv[i])=="-sol"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no filename for the solution file given!\n"));
		error=true;
	    }
	    else{
    		solFileName=std::string(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="imatfile" || std::string(argv[i])=="-imatfile"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no filename for the incidence matrix file given!\n"));
		error=true;
	    }
	    else{
    		iMatFileName=std::string(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="selfile" || std::string(argv[i])=="-selfile"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no filename for the file with preselected columns given!\n"));
		error=true;
	    }
	    else{
    		selFileName=std::string(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="pc" || std::string(argv[i])=="-pc"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no filename for the file with prescribed columns given!\n"));
		error=true;
	    }
	    else{
    		pcFileName=std::string(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="learnfrom" || std::string(argv[i])=="-learnfrom"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no filename for the learnfile given!\n"));
		error=true;
	    }
	    else{
    		learnFileName=std::string(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="n" || std::string(argv[i])=="-n"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for n given!\n"));
		error=true;
	    }
	    else{
    		n=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="-autosearchcodes" || std::string(argv[i])=="autosearchcodes"){
	    autoSearchCodes=true;
	}
	if (std::string(argv[i])=="-autosearcharcs" || std::string(argv[i])=="autosearcharcs"){
	    autoSearchArcs=true;
	}
	if (std::string(argv[i])=="upper_n" || std::string(argv[i])=="-upper_n"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for upper_n given!\n"));
		error=true;
	    }
	    else{
    		autosearchcodes_upper_n=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="d" || std::string(argv[i])=="-d"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for d given!\n"));
		error=true;
	    }
	    else{
		unsigned int val=atoi(argv[i+1]);
		if (val<0 || val>((unsigned long long)(1)<<(sizeof(WEIGHT_SUM)*8))-1){
		    writeToConsole(std::string("Error: value for d out of range!\n"));
		    quitHeurico(-1);
		}
		d=(WEIGHT_SUM)val;
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="k" || std::string(argv[i])=="-k"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for k given!\n"));
		error=true;
	    }
	    else{
		i++;
		k=atoi(argv[i]);
		if (i+k>=argc){
		    writeToConsole(std::string("Error: not enough lambdas given!\n"));
		    error=true;
		}
		else{
		    for (int j=0; j<k; j++){
			lambda.push_back(atoi(argv[i+1+j]));
		    }
		    i+=k;
		}
	    }
	    continue;
	}
	if (std::string(argv[i])=="maxhd" || std::string(argv[i])=="-maxhd"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for maxhd given!\n"));
		error=true;
	    }
	    else{
    		maxHashDepth=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="minpc" || std::string(argv[i])=="-minpc"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for minpc given!\n"));
		error=true;
	    }
	    else{
    		minPeriod_cols=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="minpm" || std::string(argv[i])=="-minpm"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for minpm given!\n"));
		error=true;
	    }
	    else{
		minPeriod_mess=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="additiveonly" || std::string(argv[i])=="-additiveonly"){
	    additiveOnly=true;
	    continue;
	}
	if (std::string(argv[i])=="homweight" || std::string(argv[i])=="-homweight"){
	    whichWeight=WEIGHT_HOMWEIGHT;
	    continue;
	}
	if (std::string(argv[i])=="hamweight" || std::string(argv[i])=="-hamweight"){
	    whichWeight=WEIGHT_HAMWEIGHT;
	    continue;
	}
	if (std::string(argv[i])=="a" || std::string(argv[i])=="-a"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for a given!\n"));
		error=true;
	    }
	    else{
    		alterBound=atof(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="ehs" || std::string(argv[i])=="-ehs"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for ehs given!\n"));
		error=true;
	    }
	    else{
		evalHashSize=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="chs" || std::string(argv[i])=="-chs"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for chs given!\n"));
		error=true;
	    }
	    else{
		columnHashSize=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="t" || std::string(argv[i])=="-t"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for t given!\n"));
		error=true;
	    }
	    else{
    		timeout=atof(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="w" || std::string(argv[i])=="-w"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for maxbranch (-w) given!\n"));
		error=true;
	    }
	    else{
    		maxBranch=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="arcs" || std::string(argv[i])=="-arcs"){
	    arcs=true;
	    if (i+1>=argc){
		writeToConsole(std::string("Error: -arcs needs dimension (k) as further argument!\n"));
		error=true;
	    }
	    else{
		k=atoi(argv[i+1]);
		i++;
	    }
	    continue;
	}
	if (std::string(argv[i])=="htmltree" || std::string(argv[i])=="-htmltree"){
	    htmlTree=true;
	    continue;
	}
	if (std::string(argv[i])=="v" || std::string(argv[i])=="-v"){
	    if (i+1>=argc){
		writeToConsole(std::string("Error: no value for verbosity given!\n"));
		error=true;
	    }
	    else{
    		outputLevel=atoi(argv[i+1])*10;
		if (outputLevel<OUTPUT_ALWAYS){
		    outputLevel=OUTPUT_ALWAYS;
		    writeToConsole(std::string("Warning: verbosity has to be >=1 and was adjusted.\n"));
		}
		else if (outputLevel>OUTPUT_ALL){
		    outputLevel=OUTPUT_ALL;
		    writeToConsole(std::string("Warning: verbosity has to be <=5 and was adjusted.\n"));
		}
		i++;
	    }
	    continue;
	}
    }

    //catch contradictorily values from calling sequence
    if (ringFileName==""){
	writeToConsole(std::string("Error: no ring file given!\n"));
	error=true;
    }
    else{
	CHAINRING::init(ringFileName.c_str());
    }
    if (n<=0){
	writeToConsole(std::string("Error: negative value for n given!\n"));
	error=true;
    }
    else if (n>MAX_LENGTH){
	writeToConsole(std::string("Error: value for n to large!\n"));
	error=true;
    }
    if (k<=0){
	writeToConsole(std::string("Error: no correct value for k given!\n"));
	error=true;
    }
    if (d==0){
	writeToConsole(std::string("Error: no correct value for d given!\n"));
	error=true;
    }
    if (n<=(int)d && autoSearchArcs){
	writeToConsole(std::string("Error: -autosearcharcs and n<=d!\n"));
	error=true;
    }
    if (!autoSearchCodes && autosearchcodes_upper_n!=MAX_LENGTH){
	writeToConsole(std::string("Error: upper_n is set while autosearchcodes is not!\n"));
	error=true;
    }
    if (arcs==false){
	if (lambda.size()==0){
	    writeToConsole(std::string("Error: no lambdas given!\n"));
	    error=true;
	}
	if (whichWeight==WEIGHT_UNDEFINED){
	    writeToConsole(std::string("Error: no weight given!\n"));
	    error=true;
	}
    }
    else{
	if (pcFileName!=""){
	    writeToConsole(std::string("Error: -pc can (currently) not be combined with -arcs!\n"));
	    error=true;
	}
	if (selFileName!=""){
	    writeToConsole(std::string("Error: -selfile can (currently) not be combined with -arcs!\n"));
	    error=true;
	}
	if (autoSearchCodes){
	    writeToConsole(std::string("Error: -autosearchcodes can not be combined with -arcs!\n"));
	    error=true;
	}
	if (autoSearchArcs){
	    writeToConsole(std::string("Error: -autosearcharcs can not be combined with -arcs!\n"));
	    error=true;
	}
	lambda.clear();
	for (int i=0; i<k; i++){
	    lambda.push_back(ChainLength<CHAINRING>()());
	}
	minPeriod_cols=ChainLength<CHAINRING>()();
	minPeriod_mess=ChainLength<CHAINRING>()();
	whichWeight=WEIGHT_HAMWEIGHT;
    }
    if (error){
        writeToConsole(std::string("usage: heurico2 -ring <ringfile> -n <length> -d <d> -k <dim, lambda_1 ... lambda_k> <-homweight || -hamweight>\n"));
	writeToConsole(std::string("optional arguments are:\n"));
	writeToConsole(std::string("-ehs <evalHashSize> -chs <columnHashSize> -maxhd <maxHashDepth> -w <maxbranch> -a <altBound>, -t <timeout>,\n"));
        writeToConsole(std::string("-minpc <minpc>, -minpm <minpm>, -sol <solfile>, -pc <pcfile>, -selfile <preselfile>, -upper_n <u_n> -htmltree.\n"));
	writeToConsole(std::string("-autosearchcodes, -autosearcharcs"));
	writeToConsole(std::string("-arcs starts a special version with lambda_1= ... =lambda_k=minpc=minpm=m, -hamweight.\n"));
	writeToConsole(std::string("-v <verbosity>\n"));
	quitHeurico(-1);
    }
    else{
	std::stringstream searchParamInfo;
	searchParamInfo<<"Starting search with the following parameters:"<<std::endl;
	searchParamInfo<<"ring="<<ringFileName<<"; n="<<n<<"; k="<<k<<";";
	for (int i=0; i<k; i++){
	    searchParamInfo<<" l"<<i<<"="<<lambda[i];
	    if (i<k-1){
		searchParamInfo<<",";
	    }
	}
	searchParamInfo<<"; d="<<(int)d<<"; ";
	switch (whichWeight){
	    case WEIGHT_HAMWEIGHT:
		searchParamInfo<<"weight = Hamming weight"<<std::endl;
		break;
	    case WEIGHT_HOMWEIGHT:
		searchParamInfo<<"weight = homogenous weight"<<std::endl;
		break;
	    default:
		searchParamInfo<<"weight = unknown weight"<<std::endl;
		break;
	}
	searchParamInfo<<"optional parameters:"<<std::endl;
	searchParamInfo<<"maxhd="<<maxHashDepth<<"; minpc="<<minPeriod_cols<<"; minpm="<<minPeriod_mess<<"; pcfile="<<pcFileName<<"; selfile="<<selFileName<<";"<<"altBound="<<alterBound<<";"<<std::endl;
	searchParamInfo<<"ehs="<<evalHashSize<<"; chs="<<columnHashSize<<"; m="<<maxBranch<<"; t="<<timeout<<"; autosearchcodes="<<autoSearchCodes<<"; arcs="<<arcs<<"; htmlTree="<<htmlTree<<std::endl;  
	writeToConsole(searchParamInfo.str());
	if (arcs==false){
	    CodewordManager<CHAINRING>* c=new CodewordManager<CHAINRING>();
	    c->setOutputLevel(outputLevel);
	    c->initializeGeometry(k,lambda,whichWeight,minPeriod_cols,minPeriod_mess,additiveOnly,iMatFileName);
	    cwm=c;
	}
	else{
	    ArcSearchManager<CHAINRING>* a=new ArcSearchManager<CHAINRING>();
	    a->setOutputLevel(outputLevel);
	    a->initializeGeometry(k,whichWeight,iMatFileName);
	    cwm=(AbstractCodewordManager<CHAINRING>*)(a);
	}
	std::vector<COLUMN_NO> pc;
	if (pcFileName!=""){
	    std::ifstream inf(pcFileName.c_str());
	    if (!inf.is_open()){
		std::stringstream error;
		error<<"Error: could not open pc-file "<<pcFileName<<"!"<<std::endl;
		writeToConsole(error.str());
		quitHeurico(-1);
	    }
	    pc=convertGMToNumberVector<CHAINRING>(inf,cwm,k);
	}
	std::vector<std::set<COLUMN_NO> > preselectedColumns;
	if (selFileName!=""){
	    std::ifstream inf(selFileName.c_str());
	    if (!inf.is_open()){
		std::stringstream error;
		error<<"Error: could not open sel-file "<<selFileName<<"!"<<std::endl;
		writeToConsole(error.str());
		quitHeurico(-1);
	    }
	    unsigned int counter=0;
	    while(true){
		std::string s;
		std::stringstream ss;
		for (unsigned int i=0; i<cwm->getMessageLength(); i++){
		    getline(inf,s);
		    if (inf.eof()){
			break;
		    }
		    if (s==""){
			i--;
			continue;
		    }
		    ss<<s<<std::endl;
		}
		if (inf.eof()){
		    break;
		}
		std::vector<COLUMN_NO> selVec=convertGMToNumberVector(ss,cwm,cwm->getMessageLength());
		std::set<COLUMN_NO> selSet;
		for (unsigned int i=0; i<selVec.size(); i++){
		    selSet.insert(selVec[i]);
		}
		preselectedColumns.push_back(selSet);
		counter++;
	    }
	    spsm.init(preselectedColumns);
	    psm=&spsm;
	}	
	std::vector<COLUMN_NO> learnVector;
	if (learnFileName!=""){
	    std::ifstream inf(learnFileName.c_str());
	    if (!inf.is_open()){
		std::stringstream error;
		error<<"Error: could not open learn-file "<<learnFileName<<"!"<<std::endl;
		writeToConsole(error.str());
		quitHeurico(-1);
	    }
	    learnVector=convertGMToNumberVector<CHAINRING>(inf,cwm,k);
	}
	if (n<=(int)pc.size()){
	    writeToConsole(std::string("Error: n <= number of prescribed columns!\n"));
	    quitHeurico(-1);
	}
	SearchAlgorithm<CHAINRING> s;
	s.setOutputLevel(outputLevel);
	s.initializeGeometry(cwm);
	if (!autoSearchCodes){
	    if (!autoSearchArcs){
    		s.initializeSearch(n,d,maxHashDepth,maxBranch,alterBound,evalHashSize,columnHashSize,solFileName,timeout,htmlTree);
		std::pair<int,bool> ret=s.startSearch(pc,learnVector,psm);
	    }
	    else{
		int u=n-d; //u>0, as n>d is checked before
		int max_n=n;
		for (int running_u=u; u<=MAX_U_ARCS; running_u++){
		    int initial_n=max_n;
		    while(true){
			std::stringstream sts1,sts2;
			sts1<<"solution_"<<running_u<<"_"<<max_n<<".txt";
			solFileName=sts1.str();
			sts2<<"\n**************************\nCalling with u="<<running_u<<", n="<<max_n<<", b="<<alterBound<<"\n";
			writeToConsole(sts2.str());
			s.initializeSearch(max_n,max_n-running_u,maxHashDepth,maxBranch,alterBound,evalHashSize,columnHashSize,solFileName,timeout,htmlTree);
			std::pair<int,bool> ret=s.startSearch(pc,learnVector,psm);
			writeToConsole(std::string("**************************\n"));
			if (ret.first==1 || ret.first==2){
			    max_n--;
			    break;
			}
			else if (ret.first==3){
			    if (max_n>initial_n){
				std::stringstream sts3;
				#ifdef LINUX
				    sts3<<"rm solution_"<<running_u<<"_"<<max_n-1<<".txt";
				#else
				    sts3<<"del solution_"<<running_u<<"_"<<max_n-1<<".txt";
				#endif
				std::system(sts3.str().c_str());
			    }
			    max_n++;
			}
		    }
		}
	    }
	}
	else{
	    if (autoSearchArcs){
		writeToConsole("Warning: -autosearcharcs was ignored.\n");
	    }
	    int last_d=d;
	    int current_min_d=d;
	    int running_n=n;
	    unsigned int default_maxBranch=maxBranch;
	    double default_alterBound=alterBound;
	    while(true){
		std::stringstream dirname,call,uboundsFileName;
		dirname<<"./";
		for (unsigned int i=0; i+1<lambda.size(); i++){
		    dirname<<lambda[i]<<"-";
		}
		dirname<<lambda[lambda.size()-1];
		uboundsFileName<<dirname.str()<<"/upperbounds";
		call<<MKDIRCALL<<" \""<<dirname.str()<<"\"";
		std::system(call.str().c_str());
		while(running_n<=autosearchcodes_upper_n){
		    std::string oldFileName;
		    current_min_d=last_d;
		    alterBound=default_alterBound;
		    maxBranch=default_maxBranch;
		    int counter=0;
		    while(true){
			std::stringstream sts1,sts2;
			sts1<<dirname.str()<<"/"<<"solution_"<<running_n<<"_"<<current_min_d<<".txt";
			solFileName=sts1.str();
			sts2<<"\n**************************\nCalling with n="<<running_n<<", d="<<current_min_d<<", a="<<alterBound<<", m = "<<maxBranch<<", maxHD = "<<maxHashDepth<<"\n";
			writeToConsole(sts2.str());
			
			std::pair<int,bool> ret;
			bool skipLongSearch=false;
			if (!skipLongSearch){
			    s.initializeSearch(running_n,current_min_d,maxHashDepth,maxBranch,alterBound,evalHashSize,columnHashSize,solFileName,timeout,htmlTree);
			    ret=s.startSearch(pc,learnVector,psm);
			}
			if (ret.first==NOSOLUTION_COMPLETE){
			    std::stringstream ss,ss2;
			    ss<<"Search completed without finding a solution."<<std::endl;
			    if (!ret.second){ //no pruning was done!
				ss<<"As no pruning was done, we have: ";
				ss2<<"dmax("<<CHAINRING::getName()<<","<<running_n<<",[";
				for (unsigned int i=0; i<lambda.size(); i++){
				    ss2<<lambda[i];
				    if (i+1<lambda.size()){
					ss2<<",";
				    }
				    else{
					ss2<<"]";
				    }
				}
				switch (whichWeight){
				    case WEIGHT_HAMWEIGHT:
				    ss2<<",ham)";
				    break;
				    case WEIGHT_HOMWEIGHT:
				    ss2<<",hom)";
				    break;
				    default:
				    ss2<<",\?\?)";
				    break;
				}
				ss2<<" <= "<<current_min_d-1<<std::endl;
				writeToConsole(ss.str()+ss2.str());
				std::ofstream uboundsFile(uboundsFileName.str().c_str(),std::ios::app);
				if (uboundsFile.is_open()){
				    uboundsFile<<ss2.str();
				    uboundsFile.close();
				}
			    }
			    writeToConsole(std::string("**************************\n"));
			    break;
			}
			else if (ret.first==NOSOLUTION_TIMEOUT){
			    writeToConsole(std::string("**************************\n"));
			    counter++;
			    switch (counter){
				case 1: maxBranch=8; alterBound=pow(alterBound,0.5); break;
				case 2: maxBranch=5; alterBound=pow(alterBound,0.5); break;
				case 3: maxBranch=3; alterBound=pow(alterBound,0.5); break;
				case 4: maxBranch=2; alterBound=pow(alterBound,0.5); break;
				case 5: current_min_d++; counter--; break;
			    }
			}
			else if (ret.first==FOUNDSOLUTION){
			    writeToConsole(std::string("**************************\n"));
			    last_d=current_min_d;
			    if (oldFileName!=""){
				std::stringstream sts3;
				#ifdef LINUX
				    sts3<<"rm "<<"\""<<oldFileName<<"\"";
				#else
				    sts3<<"del "<<"\""<<ReplaceString("/","\\",oldFileName)<<"\"";
				#endif
				std::system(sts3.str().c_str());
			    }
			    oldFileName=sts1.str();
			    current_min_d++; alterBound=default_alterBound; maxBranch=default_maxBranch; counter=0;
			}
		    }
		    running_n++;
		}
		lambda=getNextShape(lambda);
		cwm->initializeGeometry(lambda.size(),lambda,whichWeight,minPeriod_cols,minPeriod_mess,additiveOnly,iMatFileName);
		s.initializeGeometry(cwm);
		current_min_d=1;
		last_d=1;
		running_n=lambda.size();
	    }
	}
    }
    return 0;
}

template <class CHAINRING>
std::vector<COLUMN_NO> convertGMToNumberVector(std::istream& inf, AbstractCodewordManager<CHAINRING>* cwm, unsigned int length){
    std::vector<Element<typename CHAINRING::type::BaseSet> > entries;
    std::vector<COLUMN_NO> res;
    std::string temp;
    unsigned int counter=0;
    while(true){
	temp="";
	inf>>temp;
	if (temp==""){
	    break;
	}
	Element<typename CHAINRING::type::BaseSet> e;
	try{
	    e=convertStringToElement<CHAINRING>(temp);
	}
	catch(int e){
	    throw e;
	}
	entries.push_back(e);
	counter++;
    }
    if ((counter%length)!=0){
	writeToConsole(std::string("Error: number of entries in GM not divisible by row dimension!\n"));
	throw int(ERROR_CONVGM2NUMBERVEC);
    }
    unsigned int n=counter/length;
    Element<typename CHAINRING::type::BaseSet>* col=new Element<typename CHAINRING::type::BaseSet>[length];
    for (unsigned int i=0; i<n; i++){
	for (unsigned int j=0; j<length; j++){
	    col[j]=entries[j*n+i];
	}
	COLUMN_NO cno;
	try{
	    cno=cwm->convertColumnToNumber(col);
	}
	catch(int e){
	    delete[] col;
	    throw e;
	}
	res.push_back(cno);
    }
    delete[] col;
    return res;
}

template <class CHAINRING>
Element<typename CHAINRING::type::BaseSet> convertStringToElement(std::string s){
    //TO DO: adapt it for general chain rings
    for (unsigned int i=0; i<Cardinality<typename CHAINRING::type::BaseSet>()(); i++){
	if (CHAINRING::getElementName(i)==s){
	    return Element<typename CHAINRING::type::BaseSet>(i);
	}
    }
    std::stringstream ss;
    ss<<"Error in convertStringToElement: could not convert "<<s<<" to ring element!"<<std::endl;
    writeToConsole(ss.str());
    throw int(ERROR_CONVSTR2ELEMENT);
}

void quitHeurico(int r){
    exit(r);
}

std::string getHeuricoInformation(){
    return callingData.str();
}

std::vector<PERIOD> getNextShape(std::vector<PERIOD> currentShape){
    std::vector<PERIOD> res;
    PERIOD fill=currentShape[0];
    unsigned int k=0;
    for (unsigned int i=0; i<currentShape.size(); i++){
	k+=currentShape[i];
    }
    unsigned int rem=k+1;
    for (unsigned int i=currentShape.size()-1; i>0; i--){
	if (currentShape[i]>1){
	    rem=k;
	    for (unsigned int j=0; j<i; j++){
		res.push_back(currentShape[j]);
		rem-=currentShape[j];
	    }
	    res.push_back(currentShape[i]-1);
	    rem-=(currentShape[i]-1);
	    fill=currentShape[i]-1;
	    break;
	}
    }
    while(rem>=fill){
	res.push_back(fill);
	rem-=fill;
    }
    if (rem>0){
	res.push_back(rem);
    }
    return res;
}

