/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef HOMOGENOUS_WEIGHT_TABLE_HPP
#define HOMOGENOUS_WEIGHT_TABLE_HPP

#include "defs.h"
#include "abstractWeightTable.hpp"
#include "./Ringe/mathHelper.hpp"

template <class CHAINRING> class HomogenousWeightTable : public AbstractWeightTable<CHAINRING>{
    public:
    WEIGHT_TYPE get_weight(PERIOD);
    WEIGHT_TYPE max_weight();
    void init();
    std::string name();
    private:
        static WEIGHT_TYPE weight[MAX_PERIOD];
};

template<class CHAINRING>
WEIGHT_TYPE HomogenousWeightTable<CHAINRING>::weight[MAX_PERIOD];

template <class CHAINRING>
void HomogenousWeightTable<CHAINRING>::init(){
    unsigned int i;
    int q=mathHelper::pow(CHAINRING::getBasePrime(),CHAINRING::getRank());
    for (i=2; i<=CHAINRING::getChainLength(); i++){
	weight[i]=q-1;
    }
    weight[1]=q;
    weight[0]=0;
}

template <class CHAINRING>
WEIGHT_TYPE HomogenousWeightTable<CHAINRING>::get_weight(PERIOD h){
    return HomogenousWeightTable<CHAINRING>::weight[h];
}

template <class CHAINRING>
WEIGHT_TYPE HomogenousWeightTable<CHAINRING>::max_weight(){
    return ResidueFieldOrder<CHAINRING>()();
}

template <class CHAINRING>
std::string HomogenousWeightTable<CHAINRING>::name(){
    return std::string("homogenous weight");
}

#endif

