/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "htmlTreeCreator.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>

HTMLTreeCreator::HTMLTreeCreator(){
    this->directoryCreated=NULL;
}

HTMLTreeCreator::~HTMLTreeCreator(){
    if (this->directoryCreated!=NULL){
	delete this->directoryCreated;
    }	    
}

void HTMLTreeCreator::newRootNode(std::string path, NODE_NUMBER maxN, TIME_TYPE t){
    std::stringstream fullPath;
    this->currentPath=path;
    this->maxNodes=maxN;
    if (this->directoryCreated!=NULL){
	delete this->directoryCreated;
    }	    
    this->dirNumber=(this->maxNodes-1)/NODES_PER_DIRECTORY+1;
    this->directoryCreated=new bool[this->dirNumber];
    for (unsigned int i=0; i<this->dirNumber; i++){
	this->directoryCreated[i]=false;
    }
    std::string dirName=createDir(0);
    fullPath<<this->currentPath<<dirName<<"/"<<"1.html";
    std::ofstream out(fullPath.str().c_str(),std::ios::out|std::ios::app);
    if (!out.is_open()){
	std::cout<<"Error in HTMLTreeCreator::newRootNode: could not open "<<fullPath.str()<<"!"<<std::endl;
    }
    else{
	out<<t<<": root created.<br><br>"<<std::endl;
	out.close();
    }
}

void HTMLTreeCreator::addEdge(NODE_NUMBER from, NODE_NUMBER to, std::string comment, TIME_TYPE t){
    if (to>this->maxNodes){
	return;
    }
    std::stringstream fullPath1,fullPath2;
    std::string dirNameFrom=createDir(from);
    std::string dirNameTo=createDir(to);
    fullPath1<<this->currentPath<<dirNameFrom<<"/"<<from<<".html";
    fullPath2<<this->currentPath<<dirNameTo<<"/"<<to<<".html";
    std::ofstream out1(fullPath1.str().c_str(),std::ios::out|std::ios::app);
    std::ofstream out2(fullPath2.str().c_str(),std::ios::out|std::ios::app);
    if (!out1.is_open()){
	std::cout<<"Error in HTMLTreeCreator::addEdge: could not open "<<fullPath1.str()<<"!"<<std::endl;
    }
    else{
	out1<<t<<": "<<"edge to node <a href=\""<<"../"<<dirNameTo<<"/"<<to<<".html"<<"\">"<<to<<"</a> added ("<<comment<<").<br>"<<std::endl;
	out1.close();
    }
    if (!out2.is_open()){
	std::cout<<"Error in HTMLTreeCreator::addEdge: could not open "<<fullPath2.str()<<"!"<<std::endl;
    }
    else{
	out2<<t<<": "<<"edge from node <a href=\""<<"../"<<dirNameFrom<<"/"<<from<<".html"<<"\">"<<from<<"</a> added ("<<comment<<").<br><br>"<<std::endl;
	out2.close();
    }
}

void HTMLTreeCreator::addComment(NODE_NUMBER node, std::string comment, TIME_TYPE t){
    std::stringstream fullPath;
    if (node>this->maxNodes){
	return;
    }
    std::string dirName=createDir(node);
    fullPath<<this->currentPath<<dirName<<"/"<<node<<".html";
    std::ofstream out(fullPath.str().c_str(),std::ios::out|std::ios::app);
    if (!out.is_open()){
	std::cout<<"Error in HTMLTreeCreator::addComment: could not open "<<fullPath.str()<<"!"<<std::endl;
    }
    else{
	out<<t<<": "<<comment<<"<br>"<<std::endl;
	out.close();
    }
}

std::string HTMLTreeCreator::getDirName(NODE_NUMBER no){
    std::stringstream name;
    unsigned int dirNumber=no/NODES_PER_DIRECTORY;
    name<<dirNumber*NODES_PER_DIRECTORY<<"-"<<(dirNumber+1)*NODES_PER_DIRECTORY-1;
    return name.str();
}

std::string HTMLTreeCreator::createDir(NODE_NUMBER no){
    std::string name=this->getDirName(no);
    unsigned int dirNumber=no/NODES_PER_DIRECTORY;
    if (this->directoryCreated[dirNumber]==false){
        std::stringstream call;
	call<<MKDIRCALL<<" \""<<this->currentPath<<name<<"\"";
	std::system(call.str().c_str());
	this->directoryCreated[dirNumber]=true;
    }
    return name;
}
