/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "polynom.h"
#include <algorithm>
#include <sstream>

Polynom::Polynom(){
}

Polynom::Polynom(int m){
    this->maxDegree=m;
    this->coeff=new double[m+1];
}

Polynom::~Polynom(){
    delete[] coeff;
}

std::string Polynom::getAsString(){
    std::stringstream s;
    for (int i=0; i<this->maxDegree; i++){
        s<<this->coeff[i]<<" x^"<<i<<" + ";
    }
    s<<this->coeff[this->maxDegree]<<" x^"<<this->maxDegree<<" ";
    return s.str();
}

Polynom* add(Polynom const* a, Polynom const* b, int max_deg){
    Polynom const* h;
    Polynom const* l;
    if (a->maxDegree>=b->maxDegree){
        h=a;
        l=b;
    }
    else{
        h=b;
        l=a;
    }
    Polynom* res=new Polynom(std::min(h->maxDegree,max_deg));
    for (int i=0; i<=std::min(l->maxDegree,max_deg); i++){
        res->coeff[i]=h->coeff[i]+l->coeff[i];
    }
    for (int i=l->maxDegree+1; i<=std::min(h->maxDegree,max_deg); i++){
        res->coeff[i]=h->coeff[i];
    }
    return res;
}

Polynom* multiply(Polynom const* a, Polynom const* b, int max_deg){
    Polynom const* h;
    Polynom const* l;
    if (a->maxDegree>=b->maxDegree){
        h=a;
        l=b;
    }
    else{
        h=b;
        l=a;
    }

    int deg=std::min(max_deg,h->maxDegree+l->maxDegree);
    Polynom* res=new Polynom(deg);
    for (int i=0; i<=deg; i++){
        double c=0.0;
        int upperBound=std::min(i,l->maxDegree);
        for (int j=std::max(0,i-h->maxDegree); j<=upperBound; j++){
            c+=l->coeff[j]*h->coeff[i-j];
        }
        res->coeff[i]=c;
    }
    return res;
}

