/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef SOLUTION_DENSITY_TABLE_HPP
#define SOLUTION_DENSITY_TABLE_HPP

#include "abstractEvaluationTable.hpp"
#include "defs.h"
#include "helpFunctions.h"
#include <vector>
#include <map>

template<class CHAINRING>
class SolutionDensityTable : public AbstractEvaluationTable<CHAINRING>{
public:
    void initialize(MessageType<CHAINRING>&, AbstractWeightTable<CHAINRING>*, unsigned int, WEIGHT_SUM);
private:
    void computeTableRecursive(unsigned int, WEIGHT_TYPE, std::map<WEIGHT_TYPE,unsigned int>&, TwoDimArray<double>&);
};

template<class CHAINRING>
void SolutionDensityTable<CHAINRING>::initialize(MessageType<CHAINRING>& mt, AbstractWeightTable<CHAINRING>* wt, unsigned int _n, WEIGHT_SUM _d){
    TwoDimArray<double> tempEval;
    this->n=_n;
    this->d=_d;

    std::map<WEIGHT_TYPE, unsigned int> weightOccurences;
    for (PERIOD i=0; i<=CHAINRING::getChainLength(); i++){
	if (mt.occurencesOfPeriod[i]>0){
            WEIGHT_TYPE w=wt->get_weight(i);
    	    std::map<WEIGHT_TYPE,unsigned int>::iterator iter=weightOccurences.find(w);
	    if (iter!=weightOccurences.end()){
		iter->second+=mt.occurencesOfPeriod[i];
	    }
	    else{
		weightOccurences.insert(std::map<WEIGHT_TYPE,unsigned int>::value_type(w,mt.occurencesOfPeriod[i]));
	    }
	}
    }

    tempEval.init(_n+1,_d+1);
    for (unsigned int i=0; i<=_n; i++){
	for (unsigned int j=0; j<=_d; j++){
	    tempEval[i][j]=0.0;
	}
    }
    this->computeTableRecursive(0,0,weightOccurences,tempEval);
    this->eval.init(_n+1,_d+1);
    for (unsigned int i=0; i<=_n; i++){
	for (unsigned int j=0; j<=_d; j++){
	    if (tempEval[i][j]!=MINUS_INFTY_DB){
    		this->eval[i][j]=(EVAL_TYPE)(tempEval[i][j]*EVAL_SHIFT);
	    }
	    else{
		this->eval[i][j]=MINUS_INFTY_LL;
	    }
	}
    }
    this->initialized=true;
}

template<class CHAINRING>
void SolutionDensityTable<CHAINRING>::computeTableRecursive(unsigned int _n, WEIGHT_TYPE minWeight, std::map<WEIGHT_TYPE,unsigned int>& weightOccurences, TwoDimArray<double>& tempEval){
    static std::vector<std::pair<WEIGHT_TYPE,unsigned int> > path;
    static WEIGHT_SUM sum;
    static std::map<WEIGHT_TYPE,unsigned int>::iterator endIter;
    
    if (_n==0){
	sum=0;
	endIter=weightOccurences.end();
    }
    std::map<WEIGHT_TYPE,unsigned int>::iterator iter;
    double delta=1.0;
    for (unsigned int i=0; i<path.size(); i++){
	iter=weightOccurences.find(path[i].first);
	delta*=calcDistNumber(path[i].second,iter->second);
    }
    WEIGHT_SUM minimum=sum;
    if (this->d<sum){
	minimum=this->d;
    }
    tempEval[_n][minimum]+=delta;
    if (_n==this->n){
	return;
    }
    iter=weightOccurences.begin();
    while(iter!=endIter){
	if (iter->first<minWeight){
	    iter++;
	    continue;
	}	    
	sum+=iter->first;
	if (_n!=0 && iter->first==minWeight){
	    path[path.size()-1].second++;
	    computeTableRecursive(_n+1,minWeight,weightOccurences,tempEval);
	    path[path.size()-1].second--;
	}
	else{
	    path.push_back(std::pair<WEIGHT_TYPE,unsigned int>(iter->first,1));
	    computeTableRecursive(_n+1,iter->first,weightOccurences,tempEval);
	    path.pop_back();
	}
        sum-=iter->first;
	iter++;
    }
    if (_n==0){
	for (unsigned int i=0; i<=this->n; i++){
	    WEIGHT_SUM w=this->d;
	    while(true){
		if (w==0){
		    break;
		}
		tempEval[i][w-1]+=tempEval[i][w];
		w--;
	    }
	    double total=tempEval[i][0];
	    //std::cout<<"n = "<<i<<" -> total = "<<total<<std::endl;
	    for (w=0; w<=this->d; w++){
		if (tempEval[i][w]>0.0){
		    tempEval[i][w]=log(tempEval[i][w]/total);
		}
		else{
		    tempEval[i][w]=MINUS_INFTY_DB;
		}
	    }
	}
    }
}

#endif

