/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#ifndef STOCHASTIC_EVALUATION_TABLE_HPP
#define STOCHASTIC_EVALUATION_TABLE_HPP

#include "abstractEvaluationTable.hpp"
#include "defs.h"

template<class CHAINRING>
class StochasticEvaluationTable : public AbstractEvaluationTable<CHAINRING>{
public:
    void initialize(MessageType<CHAINRING>&, AbstractWeightTable<CHAINRING>*, unsigned int, WEIGHT_SUM);
};

template<class CHAINRING>
void StochasticEvaluationTable<CHAINRING>::initialize(MessageType<CHAINRING>& mt, AbstractWeightTable<CHAINRING>* wt, unsigned int _n, WEIGHT_SUM _d){
    this->n=_n;
    this->d=_d;
    TwoDimArray<double> tempEval;

    WEIGHT_TYPE max_occ_weight=0;
    unsigned int total=0;

    //compute maximum weight and total sum
    for (PERIOD i=0; i<=CHAINRING::getChainLength(); i++){
        if (mt.occurencesOfPeriod[i]>0){
            total+=mt.occurencesOfPeriod[i];
            max_occ_weight=(std::max)(max_occ_weight,wt->get_weight(i));
        }
    }

    //count how often every weight occurs
    unsigned int* weightCounter=new unsigned int[max_occ_weight+1];
    for (unsigned int i=0; i<=max_occ_weight; i++){
        weightCounter[i]=0;
    }
    for (PERIOD i=0; i<=CHAINRING::getChainLength(); i++){
        weightCounter[wt->get_weight(i)]+=mt.occurencesOfPeriod[i];
    }

    tempEval.init(_n+1,_d+1);

    //now compute probabilities
    Polynom* curr=new Polynom(0);
    Polynom* mul=new Polynom(max_occ_weight);
    curr->coeff[0]=1.0;
    for (unsigned int i=0; i<=max_occ_weight; i++){
        mul->coeff[i]=(double)(weightCounter[i])/(double)(total);
    }
    tempEval[0][0]=log(1.0);
    for (unsigned int i=1; i<=_d; i++){
        tempEval[0][i]=MINUS_INFTY_LL;
    }
    for (unsigned int i=1; i<=_n; i++){
        Polynom* temp=multiply(curr,mul);
        delete curr;
        curr=temp;
        for (unsigned int j=0; j<=_d; j++){
            tempEval[i][j]=0.0;
            for (int k=j; k<=curr->maxDegree; k++){
                tempEval[i][j]+=curr->coeff[k];
            }
            if (tempEval[i][j]>0.0){
                tempEval[i][j]=log(tempEval[i][j]);
            }
            else{
                tempEval[i][j]=MINUS_INFTY_DB;
            }
        }
    }
    this->eval.init(_n+1,_d+1);
    for (unsigned int i=0; i<=_n; i++){
	for (unsigned int j=0; j<=_d; j++){
	    if (tempEval[i][j]!=MINUS_INFTY_DB){
    		this->eval[i][j]=(EVAL_TYPE)(tempEval[i][j]*EVAL_SHIFT);
	    }
	    else{
		this->eval[i][j]=MINUS_INFTY_LL;
	    }
	}
    }
    delete curr;
    delete mul;
    delete[] weightCounter;
    this->initialized=true;
}
#endif
