/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/Log.h"
#include "../h/NetCommunicator.h"

#include <sstream>
#include <cstring>

#ifdef WIN32
	int NetCommunicator::startWinsock(){
		long rc;
		WSADATA wsa;
		rc=WSAStartup(MAKEWORD(2,0),&wsa);
		if(rc!=0){
			stringstream s;
			s<<"Fehler: startWinsock, Fehlercode: "<<rc<<std::endl;
			writeToConsole(s);
			return 1;
		}
		else{
			stringstream s;
			s<<"Winsock gestartet!"<<std::endl;
			writeToConsole(s);
		}
		return rc;
	}
#endif

void NetCommunicator::closeSocket(){
	#ifdef WIN32
		closesocket(this->itsSocket);
	#endif
	#ifdef LINUX
		close(this->itsSocket);
	#endif
}

long NetCommunicator::resolveHostname(string hostname, string& ip){
	#ifdef WIN32
		// Winsock starten
		this->startWinsock();
	#endif

	hostent* he=gethostbyname((const char*)hostname.c_str());
    if(he == NULL){
		stringstream s;
        s<<"Host konnte nicht aufgeloest werden!"<<std::endl;
		writeToConsole(s);
        return 1;
    }
	ip=string(inet_ntoa(*((in_addr*)(he->h_addr_list[0]))));
	return 0;
}	

int NetCommunicator::createSocketByIP(string ip, int port){
	long rc;
	SOCKADDR_IN addr;

	#ifdef WIN32
		// Winsock starten
		this->startWinsock();
	#endif

	// Socket erstellen
	this->itsSocket=socket(AF_INET,SOCK_STREAM,0);
	if(this->itsSocket==INVALID_SOCKET)
	{
		stringstream s;
		s<<"Fehler: Der Socket konnte nicht erstellt werden!";
		#ifdef WIN32
			s<<" Fehlercode: "<<WSAGetLastError();
		#endif
		s<<std::endl;
		writeToConsole(s);
		return 1;
	}
	else
	{
		stringstream s;
		s<<"Socket erstellt!"<<std::endl;
		writeToConsole(s);
	}

	// Verbinden
	memset(&addr,0,sizeof(SOCKADDR_IN)); // zuerst alles auf 0 setzten
	addr.sin_family=AF_INET;
	addr.sin_port=htons(port);
	addr.sin_addr.s_addr=inet_addr(ip.c_str()); // zielrechner ist unser eigener

	rc=connect(this->itsSocket,(SOCKADDR*)&addr,sizeof(SOCKADDR));
	if(rc==SOCKET_ERROR)
	{
		stringstream s;
		s<<"Fehler: connect() gescheitert!";
		#ifdef WIN32
			s<<" Fehlercode: "<<WSAGetLastError();
		#endif
		s<<std::endl;
		writeToConsole(s);
		return 1;
	}
	else
	{
		stringstream s;
		s<<"Verbunden mit "<<ip<<"..."<<std::endl;
		writeToConsole(s);
		return 0; 
	}
}

void NetCommunicator::sendMessageToServer(string message){
	::send(this->itsSocket,message.c_str(),message.length(),0);
}

string NetCommunicator::readDataFromServer(){
	char buffer[4100];
	int bytesRead;
	string data;
	while((bytesRead=::recv(this->itsSocket, buffer, 4096, 0)) > 0){
		buffer[bytesRead]=0;
		data+=string(buffer);
    }
	return data;
}

string NetCommunicator::getFileFromIP(string ip,string filename){
	long rc;
	rc=this->createSocketByIP(ip,80);
	stringstream request;
	request<<"GET "<<filename<<" HTTP/1.1\r\nHost: "<<ip<<"\r\nConnection: close\r\n\r\n";
	this->sendMessageToServer(request.str());
	string data=this->readDataFromServer();
	this->closeSocket();
	return data;
}

string NetCommunicator::getFileFromHost(string host,string filename){
	long rc;
	string ip;
	rc=this->resolveHostname(host,ip);
	return this->getFileFromIP(ip,filename);
}
