/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/Probtable.h"

ProbTable::ProbTable(){
    this->mem_allocated=false;
}
ProbTable::~ProbTable(){
    this->freeMemory();
}

void ProbTable::freeMemory(){
    if (mem_allocated){
	delete[] this->prob;
	this->mem_allocated=false;
    }
}

void ProbTable::init(int l, int u, double rv_llimit, double rv_ulimit){
    this->l_limit=l;
    this->u_limit=u;
    this->retval_llimit=rv_llimit;
    this->retval_ulimit=rv_ulimit;
    if (l<=u){
	this->prob=new MyFPType[u-l+1];
	for (int i=0; i<=this->u_limit-this->l_limit; i++){
	    this->prob[i].setValueByDouble(0.0);
	}
	this->mem_allocated=true;
    }
    else{
	this->prob=NULL;
    }
}

int ProbTable::get_llimit(){
    return this->l_limit;
}

int ProbTable::get_ulimit(){
    return this->u_limit;
}

double ProbTable::get_val(int i){
    if (i<this->l_limit){
	return retval_llimit;
    }
    if (i>this->u_limit){
	return retval_ulimit;
    }
    return this->prob[i-this->l_limit].getDoubleValue();
}
void ProbTable::set_val(int i, MyFPType d){
    if (i>=this->l_limit && i<=this->u_limit){
	this->prob[i-this->l_limit]=d;
    }
}
void ProbTable::add_val(int i, MyFPType d){
    if (i>=this->l_limit && i<=this->u_limit){
	this->prob[i-this->l_limit]+=d;
    }
}
void ProbTable::normalize(){
    MyFPType pmax=this->prob[this->u_limit-this->l_limit];
    if (pmax.getDoubleValue()>0.0){
	for (int i=this->l_limit; i<=this->u_limit; i++){
	    this->prob[i-this->l_limit]/=pmax;
	    this->prob[i-this->l_limit].expandDouble();
	}
    }
}
void ProbTable::copy(ProbTable& a){
    this->freeMemory();
    this->init(a.l_limit,a.u_limit);
    this->retval_llimit=a.retval_llimit;
    this->retval_ulimit=a.retval_ulimit;
    for (int i=this->l_limit; i<=this->u_limit; i++){
	this->prob[i-this->l_limit]=a.prob[i-this->l_limit];
    }
}

