/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/SampledProbEvaluator.h"
#include "../h/Log.h"
#include "../h/Random.h"
#include "../h/CommonRoutines.h"
#include <iostream>
#include <vector>
#include <math.h>
using namespace std;

SampledProbEvaluator::SampledProbEvaluator(){
	this->sampleRuns=DEFAULT_SAMPLE_RUNS;
	this->PAC=DEFAULT_PAC;
	this->hashAllocated=false;
	this->choiceTable=NULL;
}

bool SampledProbEvaluator::getRowIntervals(RowIntervalTable** rivt, bool* rivtAlloc){
    return false;
}

unsigned int SampledProbEvaluator::getSampleRuns(){
	return this->sampleRuns;
}

void SampledProbEvaluator::setSampleRuns(unsigned int n){
	this->sampleRuns=n;
}

double SampledProbEvaluator::getPAC(){
	return this->PAC;
}

void SampledProbEvaluator::setPAC(double d){
	this->PAC=d;
}

void SampledProbEvaluator::clearHash(){
	if (this->hashAllocated){
		unsigned int i;
		for (i=0; i<this->problem->matrix.rows; i++){
			this->hash[i].clear();
		}
		delete this->hash;
	}
}

void SampledProbEvaluator::setProblem(Problem* p){
	this->clearHash();
	this->problem=p;
	this->hash=new map<unsigned long long, EVAL_TYPE>[this->problem->matrix.rows];
	this->hashAllocated=true;
	this->choiceTable=computeChoices(p->matrix.cols,p->controlRow,p->CRLowerBound,p->CRUpperBound);
}

Problem* SampledProbEvaluator::getProblem(){
	return this->problem;
}

unsigned long long SampledProbEvaluator::computeKey(sum_of_entries rhs, unsigned int CRPhase, sum_of_entries cr){
    return ((unsigned long long)(rhs)<<32)+((unsigned long long)(CRPhase)<<16)+cr;
}

EVAL_TYPE SampledProbEvaluator::getEvaluation(unsigned int c, RowListEntry* relevantRows, unsigned int CRPhase, sum_of_entries currentCR){
    RowListEntry* rle;
    unsigned int i,j,k,l;
    int index=c*this->problem->matrix.rows;
    sum_of_entries CR_LB=this->problem->CRLowerBound;
    sum_of_entries CR_UB=this->problem->CRUpperBound;

    if (currentCR>CR_UB || (currentCR<CR_LB && this->choiceTable[CR_UB-currentCR].choices.size()==0)){
	return MINUS_INFTY_LL;
    }

    EVAL_TYPE eval=0;
    vector<RowListEntry*> rowsToSample;	

    map<unsigned long long,EVAL_TYPE>::iterator iter;
    for (rle=relevantRows->nextEntry; rle!=relevantRows; rle=rle->nextEntry){
	rle->tempRHS=rle->correspondingRHS+this->problem->matrix.data[index+rle->nr];
	unsigned long long key=this->computeKey(rle->tempRHS,CRPhase,currentCR);
	if ((iter=this->hash[rle->nr].find(key))==this->hash[rle->nr].end()){
	    rowsToSample.push_back(rle);
	}
	else{
	    if (iter->second==MINUS_INFTY_LL){
		return MINUS_INFTY_LL;
	    }
	    eval+=iter->second;
	}
    }
    unsigned int rowsToSampleSize=rowsToSample.size();

    if (rowsToSampleSize==0){
	return eval;
    }

    unsigned int* chosenColumn=new unsigned int[MAX_SAMPLE_COLUMNS];

    sum_of_entries* sum=new sum_of_entries[rowsToSampleSize];
    sum_of_entries controlSum;
    unsigned int* hits=new unsigned int[rowsToSample.size()];

    for (i=0; i<rowsToSampleSize; i++){ //initialize hitcounter with zero
	hits[i]=0;
    }

    for (i=0; i<this->getSampleRuns(); i++){ //start sampling
	controlSum=currentCR;

	j=0;
	while(true){
	    bool anotherColumn=false;
	    if (controlSum<this->problem->CRLowerBound){
		anotherColumn=true;
	    }
	    else if (controlSum<=this->problem->CRUpperBound && this->choiceTable[this->problem->CRUpperBound-controlSum].choices.size()>0){
		anotherColumn=(get_rand()<=this->getPAC());
	    }
	    if (j>=MAX_SAMPLE_COLUMNS){
		anotherColumn=false;
	    }
	    if (anotherColumn){
		while(true){
		    int choice=(unsigned int)(get_rand()*(this->choiceTable[this->problem->CRUpperBound-controlSum].choices.size()));
		    chosenColumn[j]=this->choiceTable[this->problem->CRUpperBound-controlSum].choices[choice];
		    if (this->problem->controlRow[chosenColumn[j]]<=this->problem->CRPhaseValue[CRPhase]){
			break;
		    }
		}
		controlSum+=this->problem->controlRow[chosenColumn[j]];
		j++;
	    }
	    else{
		break;
	    }
	}
	for (k=0; k<rowsToSampleSize; k++){ //initialize row- and controlcounter
	    sum[k]=rowsToSample[k]->tempRHS;
	}

	for (l=0; l<j; l++){
	    for (k=0; k<rowsToSampleSize; k++){
		sum[k]+=this->problem->matrix.data[chosenColumn[l]*this->problem->matrix.rows+rowsToSample[k]->nr];
	    }
	}

	int totalHits=0;
	for (k=0; k<rowsToSampleSize; k++){ //check for hits in the rows individually
	    if (sum[k]>=rowsToSample[k]->correspondingLowerBound && sum[k]<=rowsToSample[k]->correspondingUpperBound){
		hits[k]++;
		totalHits++;
	    }
	}
    }

    //sampling finished, now compute eval
    for (i=0; i<rowsToSampleSize; i++){
	EVAL_TYPE rowEval;
	if (hits[i]==0){
	    this->hash[rowsToSample[i]->nr].insert(map<unsigned long long,EVAL_TYPE>::value_type(this->computeKey(rowsToSample[i]->tempRHS,CRPhase,currentCR),MINUS_INFTY_LL));
	    delete[] chosenColumn;
	    delete[] sum;
	    delete[] hits;
	    return MINUS_INFTY_LL;
	}
	rowEval=(EVAL_TYPE)(log((double)(hits[i])/this->getSampleRuns())*EVAL_SHIFT);
	this->hash[rowsToSample[i]->nr].insert(map<unsigned long long,EVAL_TYPE>::value_type(this->computeKey(rowsToSample[i]->tempRHS,CRPhase,currentCR),rowEval));
	eval+=rowEval;
    }
    delete[] chosenColumn;
    delete[] sum;
    delete[] hits;
    return eval;
}
