/********************************************
Copyright 2010 by Johannes Zwanzger, Bayreuth
********************************************/

#include "../h/myFPType.h"
#include "../h/Log.h"
#include <math.h>
#include <iostream>
#include <sstream>
#include <map>

//constructors
MyFPType::MyFPType(){
}

MyFPType::MyFPType(double v){
    this->basic_value=v;
    this->ePower=0;
}

MyFPType::MyFPType(const MyFPType& v){
    (*this)=v;
}

//help functions
double MyFPType::absValue(double& a){
    if (a>0.0){
        return a;
    }
    return -a;
}

//set and get
void MyFPType::setValueByDouble(double v){
    this->basic_value=v;
    this->ePower=0;
}

double MyFPType::getDoubleValueWithExpansion(){
    if (this->ePower){
	this->expandDouble();
    }
    return this->basic_value;
}

double MyFPType::getDoubleValue() const{
    return this->basic_value;
}

//conversion
void MyFPType::normalize(){
    if (this->absValue(this->basic_value)>myFPReductionUpperMargin){
	int epow=(int)(log(this->absValue(this->basic_value)));
	double factor=exp((double)(-epow));
	this->ePower+=epow;
	this->basic_value*=factor;
    }
}

void MyFPType::changeEpow(int newpow){
    double factor=exp((double)(this->ePower-newpow));
    this->ePower=newpow;
    this->basic_value*=factor;
}

void MyFPType::expandDouble(){
    double factor=exp((double)this->ePower);
    this->basic_value*=factor;
    this->ePower=0;
}

//arithmetical single-valued operators 
MyFPType& MyFPType::operator+=(MyFPType b){
    if (this->ePower>b.ePower){
	b.changeEpow(this->ePower);
    }
    else if (this->ePower<b.ePower){
	this->changeEpow(b.ePower);
    }
    this->basic_value+=b.basic_value;
    this->normalize();
    return *this;
}

MyFPType& MyFPType::operator-=(MyFPType b){
    if (this->ePower>b.ePower){
	b.changeEpow(this->ePower);
    }
    else if (this->ePower<b.ePower){
	this->changeEpow(b.ePower);
    }
    this->basic_value-=b.basic_value;
    this->normalize();
    return *this;
}

MyFPType& MyFPType::operator*=(MyFPType b){
    this->basic_value*=b.basic_value;
    this->ePower+=b.ePower;
    this->normalize();
    return *this;
}

MyFPType& MyFPType::operator/=(MyFPType b){
    this->basic_value/=b.basic_value;
    this->ePower-=b.ePower;
    this->normalize();
    return *this;
}

//arithmetical two-valued operators 
MyFPType operator+(MyFPType a, MyFPType b){
    MyFPType res(a);
    res+=b;
    return res;
}

MyFPType operator-(MyFPType a, MyFPType b){
    MyFPType res(a);
    res-=b;
    return res;
}

MyFPType operator*(MyFPType a, MyFPType b){
    MyFPType res(a);
    res*=b;
    return res;
}

MyFPType operator/(MyFPType a, MyFPType b){
    MyFPType res(a);
    res/=b;
    return res;
}

